/*++========================================================================
 * Program Name:     Novell NCP Redirector for Linux
 * File Name:        inode.c
 * Version:          v1.00
 * Author:           James Turner
 *
 * Abstract:         This module contains functions used to control
 *                   access to the Linux file system.
 * Notes:
 * Revision History: 
 *
 *
 * Copyright (C) 2005 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *======================================================================--*/

/*===[ Include files specific to Linux ]==================================*/
#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/dcache.h>
#include <linux/mount.h>
#include <linux/pagemap.h>
#include <linux/string.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/unistd.h>
#include <linux/backing-dev.h>
#include <asm/statfs.h>
#include <asm/uaccess.h>
#include <linux/ctype.h>
#include <linux/statfs.h>

/*===[ Include files specific to this module ]============================*/
#include "vfs.h"

/*===[ External data ]====================================================*/

/*===[ External prototypes ]==============================================*/
extern int DbgPrint( char *Fmt, ... );
extern int LocalPrint( char *Fmt, ... );

extern int Init_Procfs_Interface( void );
extern void Uninit_Procfs_Interface( void );
extern void mydump(int size, void *dumpptr);

/*
 * Daemon.c functions
 */
extern void Init_Daemon_Queue( void );
extern void Uninit_Daemon_Queue( void );
extern int do_logout( struct qstr *Server );
extern int Daemon_SetMountPoint( char *Path );

/*
 * file.c functions
 */
extern int Novfs_verify_file( struct qstr *Path, session_t SessionId );
extern int Novfs_get_alltrees(struct dentry *parent);
extern int Novfs_Get_Connected_Server_List( unsigned char **ServerList, session_t SessionId );
extern int Novfs_Get_Server_Volume_List( struct qstr *Server, unsigned char **VolumeList, session_t SessionId );
extern int Novfs_Verify_Server_Name( struct qstr *Server, session_t SessionId );
extern int Novfs_Verify_Volume_Name( struct qstr *Server,  struct qstr *Volume, session_t SessionId );
extern int Novfs_Get_File_Info( unsigned char *Path, PENTRY_INFO Info, session_t SessionId  );
extern int Novfs_Get_Directory_List( unsigned char *Path, unsigned long *EnumHandle, PENTRY_INFO Info, session_t SessionId );
extern int Novfs_Get_Directory_ListEx( unsigned char *Path, unsigned long *EnumHandle, int *Count, PENTRY_INFO *Info, session_t SessionId );
extern int Novfs_Open_File( unsigned char *Path, int Flags, PENTRY_INFO Info, unsigned long *Handle, session_t SessionId );
extern int Novfs_Create( unsigned char *Path, int DirectoryFlag, session_t SessionId );
extern int Novfs_Close_File( unsigned long Handle, session_t SessionId );
extern int Novfs_Read_File( unsigned long Handle, unsigned char *Buffer, size_t *Bytes, loff_t *Offset, int User, session_t SessionId);
extern int Novfs_Write_File( unsigned long Handle, unsigned char *Buffer, size_t *Bytes, loff_t *Offset, session_t SessionId);
extern int Novfs_Delete( unsigned char *Path, int DirectoryFlag, session_t SessionId );
extern int Novfs_Truncate_File( unsigned char *Path, int PathLen, session_t SessionId );
extern int Novfs_Rename_File( int DirectoryFlag, unsigned char *OldName, int OldLen, unsigned char *NewName, int NewLen, session_t SessionId );
extern int Novfs_Set_Attr( unsigned char *Path, struct iattr *Attr, session_t SessionId );

/*
 * scope.c functions
 */
extern void Scope_Init( void );
extern void Scope_Uninit( void );
extern void *Scope_Lookup( void );
extern unsigned long Scope_Get_Hash( void *);
extern uid_t Scope_Get_Uid( void *);
extern session_t Scope_Get_SessionId( void *Scope );
extern void *Scope_Get_ScopefromName( struct qstr *Name );
extern void *Scope_Get_ScopefromPath( struct dentry *Dentry );
extern char *Scope_Get_ScopeUsers( void );
extern int Scope_Set_UserSpace( uint64_t *TotalSize, uint64_t *Free, uint64_t *TotalEnties, uint64_t *FreeEnties );
extern int Scope_Get_UserSpace( uint64_t *TotalSize, uint64_t *Free, uint64_t *TotalEnties, uint64_t *FreeEnties );
extern char *Scope_dget_path( struct dentry *Dentry, char *Buf, unsigned int Buflen, int Flags);

/*
 * profile.c functions
 */
extern int init_profile( void );

/*===[ Manifest constants ]===============================================*/
#define FILE_UPDATE_TIMEOUT   2

/*===[ Type definitions ]=================================================*/

/*===[ Function prototypes ]==============================================*/
int Novfs_Remove_from_Root(char *RemoveName);
int Novfs_Add_to_Root(char *);
char  *Novfs_dget_path( struct dentry *d, char *path, unsigned int pathlen );
int verify_dentry(struct dentry *dentry, int Flags);
int invalidate_dentry(struct dentry *parent);
struct dentry *Novfs_d_lookup(struct dentry *Parent, struct qstr *Name);
int Novfs_d_add(struct dentry *p, struct dentry *d, struct inode *i, int add);
int Novfs_d_strcmp (struct qstr *s1, struct qstr *s2);
struct inode *Novfs_get_inode(struct super_block *sb, int mode, int dev, uid_t uid, ino_t ino, struct qstr *name);
unsigned long Novfs_internal_hash (struct qstr *name);

struct super_block *Novfs_get_sb (struct file_system_type *Fstype, int Flags, const char *Dev_name, void *Data);
void Novfs_kill_sb(struct super_block *SB);
int Novfs_fill_super (struct super_block *SB, void *Data, int Silent);

/*
 * Declared dentry_operations
 */
int Novfs_d_revalidate(struct dentry *, struct nameidata *);
int Novfs_d_hash (struct dentry *, struct qstr *);
int Novfs_d_compare (struct dentry *, struct qstr *, struct qstr *);
int Novfs_d_delete(struct dentry *dentry);
void Novfs_d_release(struct dentry *dentry);
void Novfs_d_iput(struct dentry *dentry, struct inode *inode);

/*
 * Declared directory operations
 */
int Novfs_dir_open(struct inode *inode, struct file *file);
int Novfs_dir_release(struct inode *inode, struct file *file);
loff_t  Novfs_dir_lseek(struct file *file, loff_t offset, int origin);
int Novfs_dir_read(struct file *file, char *buf, size_t len, loff_t *off);
void addtodentry( struct dentry *Parent, unsigned char *List, int Level );
int Novfs_filldir(void *data, const char *name, int namelen, loff_t off, ino_t ino, unsigned ftype);
int Novfs_dir_readdir(struct file * filp, void * dirent, filldir_t filldir);
int Novfs_dir_fsync(struct file * file, struct dentry *dentry, int datasync);

/*
 * Declared address space operations
 */
int Novfs_a_readpage(struct file *file, struct page *page);

/*
 * Declared file_operations
 */
ssize_t Novfs_f_read(struct file *, char *, size_t, loff_t *);
ssize_t Novfs_f_write(struct file *, const char *, size_t, loff_t *);
int Novfs_f_readdir(struct file *, void *, filldir_t);
int Novfs_f_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int Novfs_f_mmap(struct file * file, struct vm_area_struct * vma);
int Novfs_f_open(struct inode *, struct file *);
int Novfs_f_flush(struct file *);
int Novfs_f_release(struct inode *, struct file *);
int Novfs_f_fsync(struct file *, struct dentry *, int datasync);
int Novfs_f_lock(struct file *, int, struct file_lock *);

/*
 * Declared inode_operations
 */
int Novfs_i_create(struct inode *,struct dentry *,int, struct nameidata *);
struct dentry * Novfs_i_lookup(struct inode *,struct dentry *, struct nameidata *);
int Novfs_i_mkdir(struct inode *,struct dentry *,int);
int Novfs_i_unlink(struct inode *dir, struct dentry *dentry);
int Novfs_i_rmdir(struct inode *,struct dentry *);
int Novfs_i_mknod(struct inode *,struct dentry *,int,dev_t);
int Novfs_i_rename(struct inode *, struct dentry *, struct inode *, struct dentry *);
int Novfs_i_permission(struct inode *inode, int mask);
int Novfs_i_setattr(struct dentry *, struct iattr *);
int Novfs_i_getattr(struct vfsmount *mnt, struct dentry *, struct kstat *);
int Novfs_i_revalidate(struct dentry *dentry);

void update_inode(struct inode *Inode, PENTRY_INFO Info);

/*
 * Declared super_operations
 */
void Novfs_read_inode(struct inode *inode);
void Novfs_write_inode(struct inode *inode);
int Novfs_notify_change(struct dentry *dentry, struct iattr *attr);
void Novfs_clear_inode(struct inode *inode);
int Novfs_show_options( struct seq_file *s, struct vfsmount *m );
int Novfs_statfs(struct super_block *sb, struct kstatfs *buf);

/*
 * Declared control interface functions
 */
ssize_t
Novfs_Control_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos);

ssize_t
Novfs_Control_write(struct file * file, const char * buf, size_t nbytes, loff_t *ppos);

int Novfs_Control_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

int __init init_novfs(void);
void __exit exit_novfs(void);

int Novfs_lock_inode_cache( struct inode *i );
void Novfs_unlock_inode_cache( struct inode *i );
int Novfs_enumerate_inode_cache( struct inode *i, struct list_head **iteration, ino_t *ino, PENTRY_INFO info);
int Novfs_get_entry( struct inode *i, struct qstr *name, ino_t *ino, PENTRY_INFO info);
int Novfs_get_entry_time( struct inode *i, struct qstr *name, ino_t *ino, PENTRY_INFO info, u64 *EntryTime);
int Novfs_get_remove_entry( struct inode *i, ino_t *ino, PENTRY_INFO info);
void Novfs_invalidate_inode_cache( struct inode *i );
PDIR_CACHE Novfs_lookup_inode_cache( struct inode *i, struct qstr *name, ino_t ino );
int Novfs_lookup_validate( struct inode *i, struct qstr *name, ino_t ino );
int Novfs_add_inode_entry( struct inode *i, struct qstr *name, ino_t ino, PENTRY_INFO info);
int Novfs_update_entry( struct inode *i, struct qstr *name, ino_t ino, PENTRY_INFO info);
void Novfs_remove_inode_entry( struct inode *i, struct qstr *name, ino_t ino);
void Novfs_free_invalid_entries( struct inode *i );
void Novfs_free_inode_cache( struct inode *i );

/*===[ Global variables ]=================================================*/
struct dentry_operations Novfs_dentry_operations = {
   .d_revalidate  = Novfs_d_revalidate,
   .d_hash        = Novfs_d_hash,
   .d_compare     = Novfs_d_compare,
   .d_delete      = Novfs_d_delete,
   .d_release     = Novfs_d_release,
   .d_iput        = Novfs_d_iput,
};

struct file_operations Novfs_dir_operations = {
   .open    = Novfs_dir_open,
   .release = Novfs_dir_release,
   .llseek  = Novfs_dir_lseek,
   .read    = Novfs_dir_read,
   .readdir = Novfs_dir_readdir,
   .fsync   = Novfs_dir_fsync,
};

static struct file_operations Novfs_file_operations = {
   .read    = Novfs_f_read,
   .write   = Novfs_f_write,
   .readdir = Novfs_f_readdir,
   .ioctl   = Novfs_f_ioctl,
   .mmap    = Novfs_f_mmap,
   .open    = Novfs_f_open,
   .release = Novfs_f_release,
   .fsync   = Novfs_f_fsync,
	.llseek  = generic_file_llseek,
};

static struct address_space_operations Novfs_aops = {
   .readpage   = Novfs_a_readpage,
};

static struct inode_operations Novfs_inode_operations = {
   .create     = Novfs_i_create,
   .lookup     = Novfs_i_lookup,
   .unlink     = Novfs_i_unlink,
   .mkdir      = Novfs_i_mkdir,
   .rmdir      = Novfs_i_rmdir,
   .mknod      = Novfs_i_mknod,
   .rename     = Novfs_i_rename,
   .setattr    = Novfs_i_setattr,
   .getattr    = Novfs_i_getattr,
};

static struct super_operations Novfs_ops = {
   .read_inode    = Novfs_read_inode,
   .statfs        = Novfs_statfs,
   .clear_inode   = Novfs_clear_inode,
	.drop_inode	   = generic_delete_inode,
   .show_options  = Novfs_show_options,

};

/* Not currently used
static struct file_operations Novfs_Control_operations = {
   .read    = Novfs_Control_read,
   .write   = Novfs_Control_write,
   .ioctl   = Novfs_Control_ioctl,
};
*/

ino_t Novfs_Inode_Number=1;

static struct file_system_type Novfs_fs_type = {
   .name    = "novfs", 
   .get_sb  = Novfs_get_sb,
   .kill_sb = Novfs_kill_sb,
	.owner	= THIS_MODULE,
};

struct dentry *Novfs_root=NULL;

int Novfs_Version_Major=NOVFS_VFS_MAJOR;
int Novfs_Version_Minor=NOVFS_VFS_MINOR;
int Novfs_Version_Sub=NOVFS_VFS_SUB;
int Novfs_Version_Release=NOVFS_VFS_RELEASE;

char *Novfs_CurrentMount=NULL;

DECLARE_MUTEX(InodeList_lock);

LIST_HEAD(InodeList);

unsigned long InodeCount=0, DCCount=0;
unsigned long File_update_timeout=FILE_UPDATE_TIMEOUT;
/*===[ Code ]=============================================================*/
static __inline__ void PRINT_DENTRY(const char *s, struct dentry *d)
{
   DbgPrint("%s: 0x%x\n",    s, d);
   DbgPrint("   d_count:      0x%x\n", d->d_count);
   DbgPrint("   d_lock:       0x%x\n", d->d_lock);
   DbgPrint("   d_inode:      0x%x\n", d->d_inode);
   DbgPrint("   d_lru:        0x%x\n" \
            "      next:      0x%x\n" \
            "      prev:      0x%x\n", &d->d_lru, d->d_lru.next, d->d_lru.prev);
   DbgPrint("   d_child:      0x%x\n" \
            "      next:      0x%x\n" \
            "      prev:      0x%x\n", &d->d_child, d->d_child.next, d->d_child.prev);
   DbgPrint("   d_subdirs:    0x%x\n" \
            "      next:      0x%x\n" \
            "      prev:      0x%x\n", &d->d_subdirs, d->d_subdirs.next, d->d_subdirs.prev);
   DbgPrint("   d_alias:      0x%x\n" \
            "      next:      0x%x\n" \
            "      prev:      0x%x\n", &d->d_alias, d->d_alias.next, d->d_alias.prev);
   DbgPrint("   d_time:       0x%x\n", d->d_time);
   DbgPrint("   d_op:         0x%x\n", d->d_op);
   DbgPrint("   d_sb:         0x%x\n", d->d_sb);
   DbgPrint("   d_flags:      0x%x\n", d->d_flags);
   DbgPrint("   d_mounted:    0x%x\n", d->d_mounted);
   DbgPrint("   d_fsdata:     0x%x\n", d->d_fsdata);
   DbgPrint("   d_cookie:     0x%x\n", d->d_cookie);
   DbgPrint("   d_parent:     0x%x\n", d->d_parent);
   DbgPrint("   d_name:       0x%x %.*s\n", &d->d_name, d->d_name.len, d->d_name.name);
   DbgPrint("      name:      0x%x\n" \
            "      len:       %d\n"   \
            "      hash:      0x%x\n", d->d_name.name, d->d_name.len, d->d_name.hash);
   DbgPrint("   d_hash:       0x%x\n" \
            "      next:      0x%x\n" \
            "      pprev:     0x%x\n", d->d_hash, d->d_hash.next, d->d_hash.pprev);
}

/*++======================================================================*/
int Novfs_Remove_from_Root(char *RemoveName)
/*
 *  Arguments:
 * 
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct qstr name;
   struct dentry *dentry;
   struct inode *dir;

   DbgPrint( "Novfs_Remove_from_Root: %s\n", RemoveName);
   name.len  = strlen(RemoveName);
   name.name = RemoveName;
   Novfs_d_hash(Novfs_root, &name);

   dentry = d_lookup( Novfs_root, &name);
   if (dentry)
   {
      if (dentry->d_inode && dentry->d_inode->u.generic_ip)
      {
         ((PINODE_DATA)(dentry->d_inode->u.generic_ip))->Scope = NULL;
      }
      dput(dentry);
   }

   
   dir = Novfs_root->d_inode;

   Novfs_lock_inode_cache( dir );
   Novfs_remove_inode_entry( dir, &name, 0);
   Novfs_unlock_inode_cache( dir );

   return(0);
}

/*++======================================================================*/
int Novfs_Add_to_Root(char *AddName)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct qstr name;
   struct inode *dir;
   ENTRY_INFO info;
   ino_t ino;

   DbgPrint( "Novfs_Add_to_Root: %s\n", AddName);
   name.len  = strlen(AddName);
   name.name = AddName;
   Novfs_d_hash(Novfs_root, &name);

   dir = Novfs_root->d_inode;

   Novfs_lock_inode_cache( dir );

   ino = 0;

   if ( !Novfs_lookup_inode_cache(dir, &name, 0))
   {
      info.mode = S_IFDIR | 0700;
      info.size = 0;
      info.atime = info.ctime = info.mtime = CURRENT_TIME;

      ino = (ino_t)InterlockedIncrement(&Novfs_Inode_Number);
      Novfs_add_inode_entry( dir, &name, ino, &info);
   }

   Novfs_unlock_inode_cache( dir );

   return(0);
}

/*++======================================================================*/
int Novfs_Add_to_Root2(char *AddName)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct dentry *entry;
   struct qstr name;
   struct inode *inode;
   void *scope;

   DbgPrint( "Novfs_Add_to_Root: %s\n", AddName);
   name.len  = strlen(AddName);
   name.name = AddName;

   Novfs_d_hash(Novfs_root, &name);

   entry = Novfs_d_lookup(Novfs_root, &name);
   DbgPrint( "Novfs_Add_to_Root: Novfs_d_lookup 0x%x\n", entry);
   if ( NULL == entry )
   {
      scope = Scope_Lookup();

      entry = d_alloc(Novfs_root, &name);
      DbgPrint( "Novfs_Add_to_Root: d_alloc 0x%x\n", entry );
      if (entry)
      {
         entry->d_op = &Novfs_dentry_operations;
         entry->d_time = jiffies+(File_update_timeout*HZ);
         /*
          * done in Novfs_d_add now... entry->d_fsdata = (void *)Novfs_internal_hash( &name );
          */
         inode = Novfs_get_inode(Novfs_root->d_sb, S_IFDIR | 0700, 0, Scope_Get_Uid(scope), 0, &name);
         DbgPrint( "Novfs_Add_to_Root: Inode=0x%x\n", inode);
         if (inode)
         {
            inode->i_atime = 
            inode->i_ctime = 
            inode->i_mtime = CURRENT_TIME;
            if ( !Novfs_d_add(Novfs_root, entry, inode, 1))
            {
               if (inode->u.generic_ip)
               {
                  ((PINODE_DATA)inode->u.generic_ip)->Flags = USER_INODE;
               }
               PRINT_DENTRY("After Novfs_d_add", entry);
            }
            else
            {
               dput(entry);
               iput(inode);
            }
         }
      }
   }
   else
   {
      dput(entry);
      PRINT_DENTRY("Novfs_Add_to_Root: After dput Dentry", entry);
   }
   return(0);
}

/*++======================================================================*/
char  *Novfs_dget_path( struct dentry *Dentry, char *Buf, unsigned int Buflen)
/*
 *  Arguments:   struct dentry *Dentry - starting entry
 *               char *Buf - pointer to memory buffer
 *               unsigned int Buflen - size of memory buffer
 * 
 *  Returns:     pointer to path.
 * 
 *  Abstract:    Walks the dentry chain building a path.
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   char *retval=&Buf[Buflen];
   struct dentry *p=Dentry;

   *(--retval) = '\0';
   Buflen--;

   if (!IS_ROOT(p) && !IS_ROOT(p->d_parent))
   {
      while (Buflen && !IS_ROOT(p) && !IS_ROOT(p->d_parent))
      {
         if (Buflen > p->d_name.len)
         {
            retval -= p->d_name.len;
            Buflen -= p->d_name.len;
            memcpy(retval, p->d_name.name, p->d_name.len);
            *(--retval) = '\\';
            Buflen--;
            p = p->d_parent;
         }
         else
         {
            retval = NULL;
            break;
         }
      }
   }
   else
   {
      *(--retval) = '\\';
   }



if (retval) DbgPrint("Novfs_dget_path: %s\n", retval);
   return(retval);
}

/*++======================================================================*/
int verify_dentry( struct dentry *dentry, int Flags )
/*
 *  Arguments:   struct dentry *dentry - entry to verify
 *
 *  Returns:     zero - Inode cache has been updated.  If not in the cache
 *                      then file doesn't exist.
 *               !zero - Error
 *
 *  Abstract:    This routine will verify if the file that dentry is pointing
 *               at exist and if it does it will put it in the inode cache of
 *               the parent.
 *
 *  Notes:
 *
 *  Environment: 
 *
 *========================================================================*/
{
   int retVal = -ENOENT;
   struct inode *dir;
   PENTRY_INFO info=NULL;
   PINODE_DATA id;
   session_t session;
   char *path, *list=NULL, *cp;
   ino_t ino;
   struct qstr name;
   int iLock=0;
   struct dentry *parent=NULL;
   u64 ctime;
   struct inode *inode;

   if ( IS_ROOT(dentry) )
   {
      DbgPrint("verify_dentry: Root entry\n");
      return( 0 );
   }

   if ( dentry && dentry->d_parent && 
        (dir = dentry->d_parent->d_inode) && 
        (id = dir->u.generic_ip) )
   {
      parent = dget_parent(dentry);

      info = Novfs_Malloc(sizeof(ENTRY_INFO)+PATH_LENGTH_BUFFER, GFP_KERNEL);

      if (info)
      {
         if (Novfs_lock_inode_cache( dir ))
         {
            name.len = dentry->d_name.len;
            name.name = dentry->d_name.name;
            name.hash = Novfs_internal_hash( &name );
            if ( !Novfs_get_entry_time(dir, &name, &ino, info, &ctime))
            {
               inode = dentry->d_inode;
               if ( inode && 
                    ( (inode->i_size != info->size) ||
                      (inode->i_mtime.tv_sec != info->mtime.tv_sec) ))
               {
                  ((PINODE_DATA)inode->u.generic_ip)->Flags |= UPDATE_INODE;
               }

               ctime = get_jiffies_64() - ctime;
               if ( Flags || ctime < (u64)(File_update_timeout*HZ) )
               {
                  retVal = 0;
                  Novfs_unlock_inode_cache(dir);
                  dput(parent);
                  Novfs_Free(info);
                  return(0);
               }
            }
            Novfs_unlock_inode_cache(dir);
         }

         if ( IS_ROOT(dentry->d_parent) )
         {
            session = Scope_Get_SessionId(Scope_Get_ScopefromName(&dentry->d_name));
         }
         else
         {
            session = Scope_Get_SessionId( id->Scope );
         }

         if ( !session )
         {
            id->Scope = Scope_Get_ScopefromPath(dentry);
            session = Scope_Get_SessionId( id->Scope );
         }

         ino = 0;
         retVal = 0;

         if ( IS_ROOT(dentry->d_parent) )
         {
            DbgPrint("verify_dentry: parent is Root directory\n");
            list = Scope_Get_ScopeUsers();

            iLock = Novfs_lock_inode_cache( dir );
            Novfs_invalidate_inode_cache( dir );

            if ( list )
            {
               cp = list;
               while (*cp)
               {
                  name.name = cp;
                  name.len  = strlen(cp);
                  name.hash = Novfs_internal_hash( &name );
                  cp += (name.len+1);
                  ino = 0;
                  if ( Novfs_get_entry( dir, &name, &ino, info ) )
                  {
                     info->mode = S_IFDIR | 0700;
                     info->size = 0;
                     info->atime = info->ctime = info->mtime = CURRENT_TIME;
                     ino = (ino_t)InterlockedIncrement(&Novfs_Inode_Number);
                     Novfs_add_inode_entry(dir, &name, ino, info);
                  }
               }
            }
            Novfs_free_invalid_entries( dir );
         }
         else
         {
            
            path = Novfs_dget_path(dentry, info->name, PATH_LENGTH_BUFFER);
            if (path)
            {
               if (dentry->d_name.len <= NW_MAX_PATH_LENGTH)
               {
                  name.hash = Novfs_internal_hash(&dentry->d_name);
                  name.len  = dentry->d_name.len;
                  name.name = dentry->d_name.name;


                  
                  retVal = Novfs_Get_File_Info( path, info, session );
                  if ( 0 == retVal)
                  {
                     dentry->d_time = jiffies+(File_update_timeout*HZ);
                     iLock = Novfs_lock_inode_cache( dir );
                     if ( Novfs_update_entry( dir, &name, 0, info ) )
                     {
                        if (dentry->d_inode)
                        {
                           ino = dentry->d_inode->i_ino;
                        }
                        else
                        {
                           ino = (ino_t)InterlockedIncrement(&Novfs_Inode_Number);
                        }
                        Novfs_add_inode_entry( dir, &name, ino, info);
                     }
                     if ( dentry->d_inode )
                     {
                        update_inode(dentry->d_inode, info);
                        id->Flags &= ~UPDATE_INODE;

                        dentry->d_inode->i_flags &= ~S_DEAD;
                        if (dentry->d_inode->u.generic_ip)
                        {
                           ((PINODE_DATA)dentry->d_inode->u.generic_ip)->Scope = id->Scope;
                        }
                     }
                  }
                  else if (-EINTR != retVal)
                  {
                     retVal = 0;
                     iLock = Novfs_lock_inode_cache( dir );
                     Novfs_remove_inode_entry( dir, &name, 0 );
                     if ( dentry->d_inode  && !(dentry->d_inode->i_flags & S_DEAD) )
                     {
                        dentry->d_inode->i_flags |= S_DEAD;
                        dentry->d_inode->i_size  = 0;
                        dentry->d_inode->i_atime.tv_sec  =
                        dentry->d_inode->i_atime.tv_nsec =
                        dentry->d_inode->i_ctime.tv_sec  =
                        dentry->d_inode->i_ctime.tv_nsec =
                        dentry->d_inode->i_mtime.tv_sec  =
                        dentry->d_inode->i_mtime.tv_nsec = 0;
                        dentry->d_inode->i_blocks = 0;
                        d_delete(dentry);   /* Remove from cache */
                     }
                  }
               }
               else
               {
                  retVal = -ENAMETOOLONG;
               }
            }
         }
      }
      else
      {
         retVal = -ENOMEM;
      }
      if (iLock)
      {
         Novfs_unlock_inode_cache( dir );
      }
      dput(parent);
   }

   if (list) Novfs_Free(list);
   if (info) Novfs_Free(info);

   DbgPrint("verify_dentry: return=0x%x\n", retVal);

   return(retVal);
}

/*++======================================================================*/
struct dentry *Novfs_d_lookup(struct dentry *Parent, struct qstr *Name)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment: 
 *
 *========================================================================*/
{
   return( d_lookup( Parent, Name));
}

/*++======================================================================*/
int Novfs_d_add(struct dentry *Parent, struct dentry *d, struct inode *i, int a)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment: 
 *
 *========================================================================*/
{
   void *scope;
   PINODE_DATA id=NULL;

   char *path, *buf;
   
   buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
   if (buf)
   {
      path = Novfs_dget_path(d, buf, PATH_LENGTH_BUFFER);
      if (path)
      {
         DbgPrint("Novfs_d_add: inode=0x%p ino=%d path %s\n", i, i->i_ino, path);
      }
      Novfs_Free(buf);
   }

   if ( Parent && Parent->d_inode && Parent->d_inode->u.generic_ip)
   {
      id = (PINODE_DATA)Parent->d_inode->u.generic_ip;
   }

   if (id && id->Scope)
   {
      scope = id->Scope;
   }
   else
   {
      scope = Scope_Get_ScopefromPath( d );
   }

   ((PINODE_DATA)i->u.generic_ip)->Scope = scope;

   d->d_time = jiffies+(File_update_timeout*HZ);
   if (a)
   {
      d_add(d, i);
   }
   else
   {
      d_instantiate(d, i);
   }
      
   return(0);
}

/*++======================================================================*/
int Novfs_d_revalidate(struct dentry *dentry, struct nameidata *nd)
/*
 *  Arguments:   struct dentry *dentry - pointer to dentry to revalidate.
 *               int flags - flags see fs.h
 * 
 *  Returns:     zero - dentry is not valid.  Caller will do a d_invalidate and dput.
 *               !zero - valid entry
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment: 
 *
 *========================================================================*/
{
   int retCode = 0;
   struct inode *dir;
   PINODE_DATA id;
   struct qstr name;

   UNUSED_VARIABLE( nd );

   DbgPrint("Novfs_d_revalidate: 0x%p %.*s\n" \
            "   d_count: %d\n" \
            "   d_inode: 0x%p\n",
      dentry, dentry->d_name.len, dentry->d_name.name, dentry->d_count, dentry->d_inode);

   if (IS_ROOT( dentry ))
   {
      retCode = 1;
   }
   else
   {
      if ( dentry->d_inode &&
           dentry->d_parent && 
           (dir = dentry->d_parent->d_inode) && 
           (id = dir->u.generic_ip) )
      {
         /*
          * Check timer to see if in valid time limit
          */
         if (jiffies > dentry->d_time)
         {
            /*
             * Revalidate entry
             */
            name.len = dentry->d_name.len;
            name.name = dentry->d_name.name;
            name.hash = Novfs_internal_hash(&dentry->d_name);
            dentry->d_time = 0;

            if ( 0 == verify_dentry( dentry, 0 ))
            {
               if (Novfs_lock_inode_cache( dir ))
               {
                  if (Novfs_lookup_inode_cache( dir, &name, 0))
                  {
                     dentry->d_time = jiffies + (File_update_timeout*HZ);
                     retCode = 1;
                  }
                  Novfs_unlock_inode_cache( dir );
               }
            }
         }
         else
         {
            retCode = 1;
         }
      }
   }


   if ( ( 0 == retCode ) && dentry->d_inode )
   {
      /*
       * Entry has become invalid
       */
/*      dput(dentry);
*/
   }

   DbgPrint("Novfs_d_revalidate: return 0x%x\n", retCode);

   return(retCode);
}

/*++======================================================================*/
unsigned long Novfs_internal_hash (struct qstr *name)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   unsigned long hash = 0;
   unsigned int len = name->len;
   unsigned char *c = (unsigned char *)name->name;

   while(len--)
   {
      /*
       * Lower case values for the hash.
       */
      hash = partial_name_hash(tolower(*c++), hash);
   }

   return(hash);
}

/*++======================================================================*/
int Novfs_d_hash (struct dentry *dentry, struct qstr *name)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   UNUSED_VARIABLE( dentry );

   DbgPrint("Novfs_d_hash: %.*s\n", name->len, name->name);

   name->hash = Novfs_internal_hash( name );
   
   return(0);
}

/*++======================================================================*/
int Novfs_d_strcmp (struct qstr *s1, struct qstr *s2)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = 1;
   unsigned char *str1, *str2;
   unsigned int len;

   DbgPrint("Novfs_d_strcmp: s1=%.*s s2=%.*s\n", s1->len, s1->name, s2->len, s2->name);

   if (s1->len && (s1->len == s2->len) && (s1->hash == s2->hash) )
   {
      len = s1->len;
      str1 = (unsigned char *)s1->name;
      str2 = (unsigned char *)s2->name;
      for ( retCode = 0; len-- ; str1++, str2++ )
      {
         if (*str1 != *str2)
         {
            if (tolower(*str1) != tolower(*str2))
            {
               retCode = 1;
               break;
            }
         }
      }
   }

   DbgPrint("Novfs_d_strcmp: retCode=0x%x\n", retCode);
   return(retCode);
}

/*++======================================================================*/
int Novfs_d_compare (struct dentry *parent, struct qstr *s1, struct qstr *s2)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode;

   retCode = Novfs_d_strcmp(s1, s2);

   DbgPrint("Novfs_d_compare: retCode=0x%x\n", retCode);
   return(retCode);
}

/*++======================================================================*/
int Novfs_d_delete(struct dentry *dentry)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retVal = 0;

     DbgPrint("Novfs_d_delete: 0x%p %.*s\n" \
            "   d_count: %d\n" \
            "   d_inode: 0x%p\n",
      dentry, dentry->d_name.len, dentry->d_name.name, dentry->d_count, dentry->d_inode);

   if (dentry->d_inode && (dentry->d_inode->i_flags & S_DEAD))
   {
      retVal = 1;
   }

   dentry->d_time = 0;

   return( retVal );
}

/*++======================================================================*/
void Novfs_d_release(struct dentry *dentry)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   DbgPrint("Novfs_d_release: 0x%x %.*s\n", dentry, dentry->d_name.len, dentry->d_name.name);
}

/*++======================================================================*/
void Novfs_d_iput(struct dentry *dentry, struct inode *inode)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   DbgPrint("Novfs_d_iput: Inode=0x%x Ino=%d Dentry=0x%x Name=%.*s\n", 
      inode, inode->i_ino, dentry, dentry->d_name.len, dentry->d_name.name);

   iput(inode);

}
/*++======================================================================*/
int Novfs_dir_open(struct inode *dir, struct file *file)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   char *path, *buf;
   

   DbgPrint("Novfs_dir_open: Inode 0x%x %d Name %.*s\n", dir, dir->i_ino, file->f_dentry->d_name.len, file->f_dentry->d_name.name);

   buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
   if (buf)
   {
      path = Novfs_dget_path(file->f_dentry, buf, PATH_LENGTH_BUFFER);
      if (path)
      {
         DbgPrint("Novfs_dir_open: path %s\n", path);
      }
      Novfs_Free(buf);
   }
   file->private_data = NULL;
   return(0);
}

/*++======================================================================*/
int Novfs_dir_release(struct inode *dir, struct file *file)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct inode *inode;

   DbgPrint("Novfs_dir_release: Inode 0x%x %d Name %.*s\n", dir, dir->i_ino, file->f_dentry->d_name.len, file->f_dentry->d_name.name);
   if (file->private_data)
   {
      inode = file->private_data;
      Novfs_free_inode_cache( inode );
      Novfs_Free(file->private_data);
   }

   return(0);
}

/*++======================================================================*/
loff_t  Novfs_dir_lseek(struct file *file, loff_t offset, int origin)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   return(dcache_dir_lseek(file, offset, origin));
}

/*++======================================================================*/
int Novfs_dir_read(struct file *file, char *buf, size_t len, loff_t *off)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
/*
   int rlen = 0;

   DbgPrint("Novfs_dir_readdir: dentry path %.*s buf=0x%p len=%d off=%lld\n", file->f_dentry->d_name.len, file->f_dentry->d_name.name, buf, len, *off);

   if (0 == *off)
   {
      rlen = 8;
      rlen -= copy_to_user(buf, "Testing\n", 8);
      *off += rlen;
   }
   return(rlen);
*/
   return(generic_read_dir(file, buf, len, off));
}

/*++======================================================================*/
void process_list(struct file *File, char *List, int Type, PENTRY_INFO Info, session_t SessionId)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   unsigned char *path, *buf=NULL, *cp;
   struct inode *dir = File->f_dentry->d_inode;
   struct qstr name;
   ino_t ino=0;

   buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
   path = buf;
   if (buf)
   {
      path = Novfs_dget_path(File->f_dentry, buf, PATH_LENGTH_BUFFER);
      if (path)
      {
         strcpy(buf, path);
      }
      path = buf+strlen(buf);
      *path++ = '\\';
   }

   if ( List )
   {
      cp = List;
      while (*cp)
      {
         name.name = cp;
         name.len  = strlen(cp);
         name.hash = Novfs_internal_hash( &name );
         cp += (name.len+1);

         Info->mode = S_IFDIR | 0700;
         Info->size = 0;
         Info->atime = Info->ctime = Info->mtime = CURRENT_TIME;

         if ( (USER_LIST != Type) && buf && ((int)(path-buf)+name.len < PATH_LENGTH_BUFFER) )
         {
            strcpy(path, name.name);
            Novfs_Get_File_Info(buf, Info, SessionId);
         }

         if ( Novfs_update_entry( dir, &name, ino, Info ) )
         {
            Novfs_add_inode_entry(dir, &name, (ino_t)InterlockedIncrement(&Novfs_Inode_Number), Info);
         }
      }
   }

   if (buf) Novfs_Free(buf);

}

/*++======================================================================*/
int Novfs_dir_readdir(struct file * file, void * dirent, filldir_t filldir)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   unsigned char *path, *buf=NULL;
   unsigned char *list=NULL;
   PENTRY_INFO info=NULL, pinfo, iinfo;
   int count;

   int status = -ENOMEM;
   struct inode *inode;
   session_t sessionId;
   uid_t uid;
   struct qstr name;
   struct list_head *inter;
   ino_t ino;
   int iLock=0;
   int type = 0;

   DbgPrint("Novfs_dir_readdir(%s): dentry path %.*s 0x%x\n", current->comm, file->f_dentry->d_name.len, file->f_dentry->d_name.name, file->f_dentry->d_fsdata);

   inode = file->f_dentry->d_inode;

   info = Novfs_Malloc(sizeof(ENTRY_INFO)+PATH_LENGTH_BUFFER, GFP_KERNEL);
   if (info)
   {
      if (file->f_pos)
      {
         if (file->private_data)
         {
            inode = file->private_data;
            while( !Novfs_get_remove_entry(inode, &ino, info) )
            {
               DbgPrint("Calling filedir-2 %llu with ino %d mode 0%o name %s\n", file->f_pos, ino, info->mode, info->name);
		         if ( filldir(dirent, info->name, info->namelength, file->f_pos, ino, info->mode >> 12) < 0 )
               {
                  /*
                   * put entry back on the list.
                   */
                  name.len  = info->namelength;
                  name.name = info->name;
                  name.hash = Novfs_internal_hash( &name );
                  Novfs_add_inode_entry(inode, &name, ino, info);
                  
                  break;
               }
		         file->f_pos++;
            }
         }
         Novfs_Free(info);
         return(0);
      }

      if ( inode && inode->u.generic_ip )
      {
         sessionId = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
         if (0 == sessionId)
         {
            ((PINODE_DATA)inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath(file->f_dentry);
            sessionId = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
         }
         uid = Scope_Get_Uid(((PINODE_DATA)inode->u.generic_ip)->Scope);
      }
      else
      {
         sessionId = 0;
         uid = current->euid;
      }

      if ( IS_ROOT(file->f_dentry) ||  /* Root */
           IS_ROOT(file->f_dentry->d_parent) || /* User */
           IS_ROOT(file->f_dentry->d_parent->d_parent) ) /* Server */
      {
         if ( IS_ROOT(file->f_dentry) )
         {
            DbgPrint("Novfs_dir_readdir: Root directory\n");
            list = Scope_Get_ScopeUsers();
            type = USER_LIST;
         
         } else if ( IS_ROOT(file->f_dentry->d_parent) )
         {
            DbgPrint("Novfs_dir_readdir: Parent is Root directory\n");
            Novfs_Get_Connected_Server_List( &list, sessionId );
            type = SERVER_LIST;
         }
         else
         {
            DbgPrint("Novfs_dir_readdir: Parent-Parent is Root directory\n");
            Novfs_Get_Server_Volume_List(&file->f_dentry->d_name, &list, sessionId);
            type = VOLUME_LIST;
         }

         iLock = Novfs_lock_inode_cache( inode );
         Novfs_invalidate_inode_cache( inode );

         process_list(file, list, type, info, sessionId);
      }
      else
      {
         DbgPrint("Novfs_dir_readdir: Default directory\n");
         buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
         if (buf)
         {
            path = Novfs_dget_path(file->f_dentry, buf, PATH_LENGTH_BUFFER);
            if (path)
            {
               unsigned long enumHandle=0;

               iLock = Novfs_lock_inode_cache( inode );
               Novfs_invalidate_inode_cache(inode);

               do
               {
                  /*
                   * Novfs_Get_Directory_List will return 0 when no error or -1 when
                   * the last entry is returned.  Otherwise it will return an error.
                   */
                  info->namelength = 0;
                  status = Novfs_Get_Directory_ListEx(path, &enumHandle, &count, &pinfo, sessionId);
                  DbgPrint("Novfs_dir_readdir: Novfs_Get_Directory_List return 0x%x count=%d pinfo=0x%p\n", status, count, pinfo);
                  iinfo = pinfo;
                  while ( (0 == status) && pinfo && count-- )
                  {
                     memcpy(info, pinfo, (int)((char *)&pinfo->name[pinfo->namelength]-(char *)pinfo));
                     pinfo = (PENTRY_INFO)&pinfo->name[pinfo->namelength];
                     name.len  = info->namelength;
                     name.name = info->name;
                     name.hash = Novfs_internal_hash(&name);
                     info->name[name.len] = '\0';

                     DbgPrint("Novfs_dir_readdir: Got %s\n", name.name);

                     if ( Novfs_update_entry( inode, &name, 0, info ))
                     {
                        Novfs_add_inode_entry(inode, &name, (ino_t)InterlockedIncrement(&Novfs_Inode_Number), info);
                     }
                  }

                  if (iinfo)
                  {
                     Novfs_Free(iinfo);
                  }
         
               } while ( !status );
            }
         }
      }

      if ( iLock )
      {
         Novfs_free_invalid_entries( inode );
      }

	   switch ((int) file->f_pos) 
	   {
	      case 0:
            DbgPrint("Calling filedir %llu with ino %d name .\n", file->f_pos, inode->i_ino);
		      if (filldir(dirent, ".", 1, file->f_pos, inode->i_ino, DT_DIR) < 0) 
		      {
			      break;
		      }
		      file->f_pos++;
		      /* fallthrough */
	      case 1:
            DbgPrint("Calling filedir %llu with ino %d name ..\n", file->f_pos, parent_ino(file->f_dentry));
		      if (filldir(dirent, "..", 2, file->f_pos, parent_ino(file->f_dentry), DT_DIR) < 0) 
		      {
			      break;
		      }
		      file->f_pos++;
		      /* fallthrough */

	      default:
            status = 0;
            inter = NULL;

            if ( iLock )
            {
               while( !Novfs_enumerate_inode_cache(inode, &inter, &ino, info) )
               {
                  DbgPrint("Calling filedir %llu with ino %d mode 0%o name %s\n", file->f_pos, ino, info->mode, info->name);
		            if ( filldir(dirent, info->name, info->namelength, file->f_pos, ino, info->mode >> 12) < 0 )
                  {
                     /*
                      * Get the rest of the entries and put them in a fake
                      * inode cache.
                      */
                     file->private_data = (PDIR_CACHE)Novfs_Malloc(sizeof(struct inode) + sizeof(DIR_CACHE), GFP_KERNEL);
                     if (file->private_data)
                     {
                        struct inode *dinode = file->private_data;
                        PINODE_DATA id = (PINODE_DATA)((char *)file->private_data+sizeof(struct inode));

                        dinode->u.generic_ip = id;

                        id->Scope = ((PINODE_DATA)inode->u.generic_ip)->Scope;
                        id->Flags = 0;
                        INIT_LIST_HEAD( &id->DirCache );
                        init_MUTEX( &id->DirCacheLock );

                        /*
                         * Copy rest of the files.
                         */
                        do
                        {
                           name.len  = info->namelength;
                           name.name = info->name;
                           name.hash = Novfs_internal_hash( &name );
                           Novfs_add_inode_entry(dinode, &name, ino, info);
                        } while( !Novfs_enumerate_inode_cache(inode, &inter, &ino, info) );
                     }

			            break;
                  }
		            file->f_pos++;
               }
            }

      }

      if (iLock)
      {
         Novfs_unlock_inode_cache(inode);
      }
   }

   if (info) Novfs_Free( info );
   if (buf)  Novfs_Free( buf );
   if (list) Novfs_Free( list );

   return(status);
}


/*++======================================================================*/
int Novfs_dir_fsync(struct file * file, struct dentry *dentry, int datasync)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   return(simple_sync_file(file, dentry, datasync));
}


/*++======================================================================*/
ssize_t Novfs_f_read(struct file *file, char *buf, size_t len, loff_t *off)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   size_t thisread, totalread=0;
   loff_t offset = *off;
   struct inode *inode;
   session_t session;
   
   if (file->f_dentry && file->f_dentry->d_inode && file->f_dentry->d_inode->u.generic_ip)
   {
      DbgPrint("Novfs_f_read(0x%x 0x%p %d %lld %.*s)\n", (unsigned long)file->private_data, buf, len, offset, file->f_dentry->d_name.len, file->f_dentry->d_name.name);
      inode = file->f_dentry->d_inode;
      
      session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
      if (0 == session)
      {
         ((PINODE_DATA)inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( file->f_dentry );
         session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
      }

      while(len > 0 && (offset < inode->i_size) )
      {
         thisread = len;
         if (Novfs_Read_File((unsigned long)file->private_data, buf, &thisread, &offset, 1, session) || !thisread)
         {
            break;
         }
         DbgPrint("Novfs_f_read thisread = 0x%x\n", thisread);
         len -= thisread;
         buf += thisread;
         offset += thisread;
         totalread += thisread;
      }
   }
   *off = offset;
   DbgPrint("Novfs_f_read return = 0x%x\n", totalread);

   return(totalread);
}

/*++======================================================================*/
ssize_t Novfs_f_write(struct file *file, const char *buf, size_t len, loff_t *off)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   size_t thiswrite, totalwrite=0;
   loff_t offset = *off;
   session_t session;
   struct inode *inode;
   int status;

   if (file->f_dentry && file->f_dentry->d_inode && file->f_dentry->d_inode->u.generic_ip)
   {
      inode = file->f_dentry->d_inode;

      DbgPrint("Novfs_f_write(0x%x %d %lld %.*s)\n", (unsigned long)file->private_data, len, offset, file->f_dentry->d_name.len, file->f_dentry->d_name.name);

      if (file->f_flags & O_APPEND) {
         offset = inode->i_size;
         DbgPrint("Novfs_f_write appending to end %lld %.*s\n", offset, file->f_dentry->d_name.len, file->f_dentry->d_name.name);
      }


      session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
      if (0 == session)
      {
         ((PINODE_DATA)inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( file->f_dentry );
         session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
      }
   
      while(len > 0)
      {
         thiswrite = len;
         if ((status = Novfs_Write_File(
               (unsigned long)file->private_data, 
               (unsigned char *)buf, 
               &thiswrite, 
               &offset, 
               session)) || !thiswrite)
         {
            totalwrite = status;
            break;
         }
         DbgPrint("Novfs_f_write thiswrite = 0x%x\n", thiswrite);
         len -= thiswrite;
         buf += thiswrite;
         offset += thiswrite;
         totalwrite += thiswrite;
         if (offset > inode->i_size)
         {
            inode->i_size = offset;
            inode->i_blocks = (offset + inode->i_blksize - 1) >> inode->i_blkbits;
         }
         inode->i_mtime = inode->i_atime = CURRENT_TIME;
         ((PINODE_DATA)inode->u.generic_ip)->Flags |= UPDATE_INODE;

      }
   }
   *off = offset;
   DbgPrint("Novfs_f_write return = 0x%x\n", totalwrite);

   return(totalwrite);
}

/*++======================================================================*/
int Novfs_f_readdir(struct file *file, void *data, filldir_t fill)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(data);
   UNUSED_VARIABLE(fill);
   return(-EISDIR);
}

/*++======================================================================*/
int Novfs_f_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   UNUSED_VARIABLE(inode);
   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(cmd);
   UNUSED_VARIABLE(arg);
   DbgPrint("Novfs_f_ioctl: file=0x%x cmd=0x%x arg=0x%x", file, cmd, arg);

   return(-ENOSYS);
}

/*++======================================================================*/
int Novfs_f_mmap(struct file * file, struct vm_area_struct * vma)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -EINVAL;

   DbgPrint("Novfs_f_mmap: file=0x%x", file);

   retCode = generic_file_mmap(file, vma);

   DbgPrint("Novfs_f_mmap: retCode=0x%x\n", retCode);
   return( retCode );
}

/*++======================================================================*/
int Novfs_f_open(struct inode *inode, struct file *file)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PENTRY_INFO info=NULL;
   int retCode = -ENOENT;
   session_t session;
   char *path;
   struct dentry *parent;
   ino_t ino;

   DbgPrint("Novfs_f_open: inode=0x%x file=0x%x dentry=0x%x dentry->d_inode=0x%x\n", inode, file, file->f_dentry, file->f_dentry->d_inode);
   if (file->f_dentry)
   {
      DbgPrint("Novfs_f_open: %.*s f_flags=0%o f_mode=0%o i_mode=0%o\n", file->f_dentry->d_name.len, file->f_dentry->d_name.name, file->f_flags, file->f_mode, inode->i_mode);
   }

   if (inode && inode->u.generic_ip)
   {
      session = Scope_Get_SessionId(((PINODE_DATA)file->f_dentry->d_inode->u.generic_ip)->Scope);
      if (0 == session)
      {
         ((PINODE_DATA)file->f_dentry->d_inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( file->f_dentry );
         session = Scope_Get_SessionId(((PINODE_DATA)file->f_dentry->d_inode->u.generic_ip)->Scope);
      }

      info = (PENTRY_INFO)Novfs_Malloc(sizeof(ENTRY_INFO) + PATH_LENGTH_BUFFER, GFP_KERNEL);
      if (info)
      {
         path = Novfs_dget_path(file->f_dentry, info->name, PATH_LENGTH_BUFFER);
         if (path)
         {
            DbgPrint("Novfs_f_open: %s\n", path);
            retCode = Novfs_Open_File(
                        path, 
                        file->f_flags & ~O_EXCL, info, 
                        (unsigned long *)&file->private_data,
                        session);
            if ( !retCode )
            {
               /*
                *update_inode(inode, &info);
                */
               parent = dget_parent(file->f_dentry);

               if (parent && parent->d_inode)
               {
                  struct inode *dir = parent->d_inode;
                  Novfs_lock_inode_cache(dir);
                  ino = 0;
                  if (Novfs_get_entry(dir, &file->f_dentry->d_name, &ino, info))
                  {
                     ((PINODE_DATA)inode->u.generic_ip)->Flags |= UPDATE_INODE;
                  }

                  Novfs_unlock_inode_cache(dir);
               }
               dput(parent);
            }
         }
         Novfs_Free(info);
      }
   }
   DbgPrint("Novfs_f_open: retCode=0x%x\n", retCode);
   return( retCode );
}

/*++======================================================================*/
int Novfs_f_flush(struct file *file)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct inode *inode;

   UNUSED_VARIABLE(file);
   if (file->f_dentry)
   {
      inode = file->f_dentry->d_inode;
      DbgPrint("Novfs_f_flush: %.*s f_flags=0x%x f_mode=0%o i_mode=0%o\n", file->f_dentry->d_name.len, file->f_dentry->d_name.name, file->f_flags, file->f_mode, inode->i_mode);
   }
   return(0);
}

/*++======================================================================*/
int Novfs_f_release(struct inode *inode, struct file *file)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -EACCES;
   session_t session;
  
   DbgPrint("Novfs_f_release: path=%.*s handle=%x\n", 
      file->f_dentry->d_name.len, 
      file->f_dentry->d_name.name, 
      (unsigned long)file->private_data);

   if ( file->f_dentry->d_inode && file->f_dentry->d_inode->u.generic_ip)
   {
      session = Scope_Get_SessionId(((PINODE_DATA)file->f_dentry->d_inode->u.generic_ip)->Scope);
      if (0 == session)
      {
         ((PINODE_DATA)file->f_dentry->d_inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( file->f_dentry );
         session = Scope_Get_SessionId(((PINODE_DATA)file->f_dentry->d_inode->u.generic_ip)->Scope);
      }

      retCode = Novfs_Close_File((unsigned long)file->private_data, session);
   }
   return(retCode);
}

/*++======================================================================*/
int Novfs_f_fsync(struct file *file, struct dentry *dentry, int datasync)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(dentry);
   UNUSED_VARIABLE(datasync);
   return(0);
}

/*++======================================================================*/
int Novfs_f_llseek(struct file *file, loff_t offset, int origin)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   DbgPrint("Novfs_f_llseek: File=0x%x Name=%.*s offset=%lld origin=%d\n", file, file->f_dentry->d_name.len, file->f_dentry->d_name.name, offset, origin);
   return(generic_file_llseek(file, offset, origin));
}

/*++======================================================================*/
int Novfs_f_lock(struct file *file, int cmd, struct file_lock *lock)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(cmd);
   UNUSED_VARIABLE(lock);
   return(-ENOSYS);
}

/*++======================================================================*/
int Novfs_a_readpage(struct file *file, struct page *page)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = 0;
   void *pbuf;
   struct inode *inode=NULL;
   struct dentry *dentry=NULL;
   loff_t offset;
   size_t len;
   session_t session=0;

   DbgPrint("Novfs_a_readpage: File=0x%x Name=%.*s Page=0x%x", file, file->f_dentry->d_name.len, file->f_dentry->d_name.name, page);

   dentry = file->f_dentry;

   if (dentry)
   {
      DbgPrint(" Dentry=0x%x Name=%.*s", dentry, dentry->d_name.len, dentry->d_name.name);
      if (dentry->d_inode)
      {
         inode = dentry->d_inode;
      }
   }



   if (inode)
   {
      DbgPrint(" Inode=0x%x Ino=%d", inode, inode->i_ino);

      if (inode->u.generic_ip)
      {
         session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
         if (0 == session)
         {
            ((PINODE_DATA)inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( file->f_dentry );
            session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
         }
      }
   }

   DbgPrint("\n");


   if (!PageUptodate(page)) 
   {
      pbuf = kmap(page);

      offset = page->index << PAGE_CACHE_SHIFT;
      len = PAGE_CACHE_SIZE;

      retCode = Novfs_Read_File((unsigned long)file->private_data, pbuf, &len, &offset, 0, session);
      if ( len && (len < PAGE_CACHE_SIZE) ) 
      {
         memset(&((char *)pbuf)[len], 0, PAGE_CACHE_SIZE-len);
      }
   
      kunmap(page);
      flush_dcache_page(page);
      SetPageUptodate(page);
   }
   unlock_page(page);

   DbgPrint("Novfs_a_readpage: retCode=%d\n", retCode);
   return (retCode);

}

/*++======================================================================*/
int Novfs_i_create(struct inode *dir, struct dentry *dentry, int mode, struct nameidata *nd)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   char *path, *buf;
   ENTRY_INFO info;
   unsigned long handle;
   session_t session;
   int retCode = -EACCES;


   DbgPrint("Novfs_i_create: mode=0%o flags=0%o %.*s\n", mode, nd->NDOPENFLAGS, dentry->d_name.len, dentry->d_name.name);

   if ( IS_ROOT(dentry) || /* Root */
        IS_ROOT(dentry->d_parent) || /* User */
        IS_ROOT(dentry->d_parent->d_parent) || /* Server */
        IS_ROOT(dentry->d_parent->d_parent->d_parent) ) /* Volume */
   {
      return(-EACCES);
   }

   if (mode | S_IFREG)
   {
      if (dir->u.generic_ip)
      {
         session = Scope_Get_SessionId( ((PINODE_DATA)dir->u.generic_ip)->Scope);
         if (0 == session)
         {
            ((PINODE_DATA)dir->u.generic_ip)->Scope = Scope_Get_ScopefromPath( dentry );
            session = Scope_Get_SessionId(((PINODE_DATA)dir->u.generic_ip)->Scope);
         }

         buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
         if (buf)
         {
            path = Novfs_dget_path(dentry, buf, PATH_LENGTH_BUFFER);
            if (path)
            {
               retCode = Novfs_Open_File(path, nd->NDOPENFLAGS|O_RDWR, &info, &handle, session);
               if ( !retCode && handle )
               {
                  Novfs_Close_File(handle, session);
                  if (!Novfs_i_mknod(dir, dentry, mode | S_IFREG, 0))
                  {
                     if (dentry->d_inode)
                     {
                        ((PINODE_DATA)dentry->d_inode->u.generic_ip)->Flags |= UPDATE_INODE;
                     }
                  }
               }
            }
            Novfs_Free(buf);
         }
      }
   }
   return( retCode );
}

/*++======================================================================*/
void update_inode(struct inode *Inode, PENTRY_INFO Info)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   static char dbuf[128];

   DbgPrint("update_inode: Inode=0x%x I_ino=%d\n", Inode, Inode->i_ino);

   DbgPrint("update_inode: atime=%s\n", ctime_r(&Info->atime.tv_sec, dbuf));
   DbgPrint("update_inode: ctime=%s\n", ctime_r(&Info->ctime.tv_sec, dbuf));
   DbgPrint("update_inode: mtime=%s\n", ctime_r(&Info->mtime.tv_sec, dbuf));
   DbgPrint("update_inode: size=%lld\n", Info->size);
   DbgPrint("update_inode: mode=0%o\n", Info->mode);

   Inode->i_mode  = Info->mode;
   Inode->i_size  = Info->size;
   Inode->i_atime = Info->atime;
   Inode->i_ctime = Info->ctime;
   Inode->i_mtime = Info->mtime;

   if (Inode->i_size && Inode->i_blksize)
   {
      Inode->i_blocks = (u_long)(Info->size >> (loff_t)Inode->i_blkbits);
      Inode->i_bytes = Info->size & (Inode->i_blksize - 1);

      DbgPrint("update_inode: i_blksize=%d\n", Inode->i_blksize);
      DbgPrint("update_inode: i_blkbits=%d\n", Inode->i_blkbits);
      DbgPrint("update_inode: i_blocks=%d\n", Inode->i_blocks);
      DbgPrint("update_inode: i_bytes=%d\n", Inode->i_bytes);
   }
}

/*++======================================================================*/
struct dentry * Novfs_i_lookup(struct inode *dir, struct dentry *dentry, struct nameidata *nd)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct dentry *retVal = ERR_PTR(-ENOENT);
   struct dentry *parent;
   PENTRY_INFO info=NULL;
   PINODE_DATA id;
   struct inode *inode=NULL;
   uid_t uid=current->euid;
   ino_t ino=0;
   struct qstr name;

   DbgPrint("Novfs_i_lookup: dir 0x%x %d name %.*s hash %d inode 0x%0p\n", dir, dir->i_ino, dentry->d_name.len, dentry->d_name.name, dentry->d_name.hash, dentry->d_inode);

   if (dir && (id = dir->u.generic_ip) )
   {
      retVal = 0;
      if ( IS_ROOT( dentry ))
      {
         DbgPrint("Novfs_i_lookup: Root entry=0x%x\n", Novfs_root);
         inode = Novfs_root->d_inode;
         return(0);
      }
      else
      {
         info = Novfs_Malloc(sizeof(ENTRY_INFO)+PATH_LENGTH_BUFFER, GFP_KERNEL);
         if (info)
         {
            if ( NULL == (retVal = ERR_PTR(verify_dentry( dentry, 1 ))))
            {
               name.name = dentry->d_name.name;
               name.len  = dentry->d_name.len;
               name.hash = Novfs_internal_hash(&name);
         
               if (Novfs_lock_inode_cache( dir ))
               {
                  if ( !Novfs_get_entry(dir, &name, &ino, info) )
                  {
                     inode = ilookup(dentry->d_sb, ino);
                     if ( inode )
                     {
                        update_inode(inode, info);
                     }
                  }
                  Novfs_unlock_inode_cache( dir );
               }

               if ( !inode && ino )
               {
                  uid = Scope_Get_Uid( id->Scope );
                  if (Novfs_lock_inode_cache( dir ))
                  {
                     inode = Novfs_get_inode(dentry->d_sb, info->mode, 0, uid, ino, &name);
                     if ( inode )
                     {
                        if ( !Novfs_get_entry(dir, &dentry->d_name, &ino, info) )
                        {
                           update_inode(inode, info);
                        }
                     }
                     Novfs_unlock_inode_cache( dir );
                  }
               }
            }
         }
      }
   }

   if ( !retVal )
   {
      dentry->d_op = &Novfs_dentry_operations;
      if (inode)
      {
         parent = dget_parent(dentry);
         Novfs_d_add(dentry->d_parent, dentry, inode, 1);
         dput(parent);
      }
      else
      {
         d_add( dentry, inode);
      }
   }

   if (info) Novfs_Free(info);

   DbgPrint("Novfs_i_lookup: inode=0x%x dentry->d_inode=0x%x return=0x%x\n", dir, dentry->d_inode, retVal);

   return(retVal);
}

/*++======================================================================*/
int Novfs_i_unlink(struct inode *dir, struct dentry *dentry)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -ENOENT;
   struct inode *inode;
   session_t session;
   char *path, *buf;
   uint64_t t64;

   DbgPrint("Novfs_i_unlink: dir=0x%x dir->i_ino=%d %.*s\n", dir, dir->i_ino, dentry->d_name.len, dentry->d_name.name);
   DbgPrint("Novfs_i_unlink: IS_ROOT(dentry)=%d\n", IS_ROOT(dentry));
   DbgPrint("Novfs_i_unlink: IS_ROOT(dentry->d_parent)=%d\n", IS_ROOT(dentry->d_parent));
   DbgPrint("Novfs_i_unlink: IS_ROOT(dentry->d_parent->d_parent)=%d\n", IS_ROOT(dentry->d_parent->d_parent));
   DbgPrint("Novfs_i_unlink: IS_ROOT(dentry->d_parent->d_parent->d_parent)=%d\n", IS_ROOT(dentry->d_parent->d_parent->d_parent) );

   if ( IS_ROOT(dentry) || /* Root */
        IS_ROOT(dentry->d_parent) || /* User */
        (!IS_ROOT(dentry->d_parent->d_parent) && /* Server */
        IS_ROOT(dentry->d_parent->d_parent->d_parent)) ) /* Volume */
   {
      return(-EACCES);
   }

   inode = dentry->d_inode;
   if ( inode )
   {
      DbgPrint("Novfs_i_unlink: dir=0x%x dir->i_ino=%d inode=0x%x ino=%d\n", dir, dir->i_ino, inode, inode->i_ino);
      if (inode->u.generic_ip)
      {
         session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
         if (0 == session)
         {
            ((PINODE_DATA)inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( dentry );
            session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
         }

         buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
         if (buf)
         {
            path = Novfs_dget_path(dentry, buf, PATH_LENGTH_BUFFER);
            if (path)
            {
               DbgPrint("Novfs_i_unlink: path %s mode 0%o\n", path, inode->i_mode);
               if (IS_ROOT(dentry->d_parent->d_parent))
               {
                  retCode = do_logout(&dentry->d_name);
               }
               else
               {
                  retCode = Novfs_Delete(path, S_ISDIR(inode->i_mode), session);
               }
               if ( !retCode || IS_DEADDIR(inode))
               {
                  Novfs_remove_inode_entry(dir, &dentry->d_name, 0);
                  dentry->d_time = 0;
                  t64 = 0;
                  Scope_Set_UserSpace(&t64, &t64, &t64, &t64);
                  retCode = 0;
               }
            }
            Novfs_Free(buf);
         }
      }
   }

   DbgPrint("Novfs_i_unlink: retCode 0x%x\n", retCode);
   return (retCode);
}

/*++======================================================================*/
int Novfs_i_mkdir(struct inode *dir, struct dentry *dentry, int mode)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   char *path, *buf;
   session_t session;
   int retCode=0;
   struct inode *inode;
   ENTRY_INFO info;
   uid_t uid;

   DbgPrint("Novfs_i_mkdir: dir=0x%p ino=%d dentry=0x%p %.*s mode=0%o\n", dir, dir->i_ino, dentry, dentry->d_name.len, dentry->d_name.name, mode);

   if ( IS_ROOT(dentry) || /* Root */
        IS_ROOT(dentry->d_parent) || /* User */
        IS_ROOT(dentry->d_parent->d_parent) || /* Server */
        IS_ROOT(dentry->d_parent->d_parent->d_parent) ) /* Volume */
   {
      return(-EACCES);
   }

   mode |= S_IFDIR;
   mode &= (S_IFMT | S_IRWXU);
   if ( dir->u.generic_ip )
   {
      session = Scope_Get_SessionId( ((PINODE_DATA)dir->u.generic_ip)->Scope);
      if (0 == session)
      {
         ((PINODE_DATA)dir->u.generic_ip)->Scope = Scope_Get_ScopefromPath( dentry );
         session = Scope_Get_SessionId(((PINODE_DATA)dir->u.generic_ip)->Scope);
      }

      uid = Scope_Get_Uid( ((PINODE_DATA)dir->u.generic_ip)->Scope);
      buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
      if (buf)
      {
         path = Novfs_dget_path(dentry, buf, PATH_LENGTH_BUFFER);
         if (path)
         {
            DbgPrint("Novfs_i_mkdir: path %s\n", path);
            retCode = Novfs_Create(path, S_ISDIR(mode), session);
            if ( !retCode )
            {
                  retCode = Novfs_Get_File_Info(path, &info, session );
                  if ( !retCode )
                  {
                     retCode = Novfs_i_mknod(dir, dentry, mode, 0);
                     inode = dentry->d_inode;
                     if (inode)
                     {
                        update_inode(inode, &info);
                        ((PINODE_DATA)inode->u.generic_ip)->Flags &= ~UPDATE_INODE;

                        dentry->d_time = jiffies+(File_update_timeout*HZ);

                        Novfs_lock_inode_cache(dir);
                        if (Novfs_update_entry( dir, &dentry->d_name, 0, &info ))
                        {
                           Novfs_add_inode_entry(dir, &dentry->d_name, inode->i_ino, &info);
                        }
                        Novfs_unlock_inode_cache(dir);
                     }

                  }
               }
         }
         Novfs_Free(buf);
      }
   }

   return( retCode );
}

/*++======================================================================*/
int Novfs_i_rmdir(struct inode *inode, struct dentry *dentry)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   return(Novfs_i_unlink(inode, dentry));
}

/*++======================================================================*/
int Novfs_i_mknod(struct inode *dir, struct dentry *dentry, int mode, dev_t dev)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct inode *inode=NULL;
   int retCode = -EACCES;
   uid_t uid;
   struct dentry *parent;

   if ( IS_ROOT(dentry) || /* Root */
        IS_ROOT(dentry->d_parent) || /* User */
        IS_ROOT(dentry->d_parent->d_parent) || /* Server */
        IS_ROOT(dentry->d_parent->d_parent->d_parent) ) /* Volume */
   {
      return(-EACCES);
   }

   if ( ((PINODE_DATA)dir->u.generic_ip) )
   {
      uid = Scope_Get_Uid( ((PINODE_DATA)dir->u.generic_ip)->Scope);
      if (mode & (S_IFREG | S_IFDIR))
      {
         inode = Novfs_get_inode(dir->i_sb, mode, dev, uid, 0, &dentry->d_name);
      }
   }

   if (inode) 
   {
      ENTRY_INFO info;

      dentry->d_op = &Novfs_dentry_operations;
      parent = dget_parent(dentry);
      Novfs_d_add(parent, dentry, inode, 0);
      memset(&info, 0, sizeof(info));
      info.mode = inode->i_mode;
      Novfs_lock_inode_cache( dir );
      Novfs_add_inode_entry( dir, &dentry->d_name, inode->i_ino, &info);
      Novfs_unlock_inode_cache( dir );

      dput(parent);
      
      retCode = 0;
   }
   DbgPrint("Novfs_i_mknod: return 0x%x\n", retCode);
   return retCode;
}

/*++======================================================================*/
int Novfs_i_rename(struct inode *odir, struct dentry *od, struct inode *ndir, struct dentry *nd)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = -ENOTEMPTY;
   char *newpath, *newbuf, *newcon;
   char *oldpath, *oldbuf, *oldcon;
   struct qstr newname, oldname;
   PENTRY_INFO info=NULL;
   int oldlen, newlen;
   session_t session;
   ino_t ino;


   if ( IS_ROOT(od) || /* Root */
        IS_ROOT(od->d_parent) || /* User */
        IS_ROOT(od->d_parent->d_parent) || /* Server */
        IS_ROOT(od->d_parent->d_parent->d_parent) ) /* Volume */
   {
      return(-EACCES);
   }

   DbgPrint("Novfs_i_rename: odir=0x%p ino=%d ndir=0x%p ino=%d\n", odir, odir->i_ino, ndir, ndir->i_ino);

   oldbuf = Novfs_Malloc(PATH_LENGTH_BUFFER*2, GFP_KERNEL);
   newbuf = oldbuf+PATH_LENGTH_BUFFER;
   if ( oldbuf && newbuf )
   {
      oldpath = Novfs_dget_path(od, oldbuf, PATH_LENGTH_BUFFER);
      newpath = Novfs_dget_path(nd, newbuf, PATH_LENGTH_BUFFER);
      if (oldpath && newpath)
      {
         oldlen = PATH_LENGTH_BUFFER-(int)(oldpath-oldbuf);
         newlen = PATH_LENGTH_BUFFER-(int)(newpath-newbuf);

         DbgPrint("Novfs_i_rename: odir=0x%x od->inode=0x%x\n", odir, od->d_inode);
         DbgPrint("Novfs_i_rename: old path %s\n", oldpath);
         DbgPrint("Novfs_i_rename: new path %s\n", newpath);

         /*
          * Check to see if two different servers or different volumes
          */
         newcon = strchr(newpath+1, '\\');
         oldcon = strchr(oldpath+1, '\\');
         DbgPrint("Novfs_i_rename: newcon=0x%x newpath=0x%x\n", newcon, newpath);
         DbgPrint("Novfs_i_rename: oldcon=0x%x oldpath=0x%x\n", oldcon, oldpath);
         retCode = -EXDEV;
         if ( newcon && oldcon && ((int)(newcon-newpath) == (int)(oldcon-oldpath)) )
         {
            newcon = strchr(newcon+1, '\\');
            oldcon = strchr(oldcon+1, '\\');
            DbgPrint("Novfs_i_rename2: newcon=0x%x newpath=0x%x\n", newcon, newpath);
            DbgPrint("Novfs_i_rename2: oldcon=0x%x oldpath=0x%x\n", oldcon, oldpath);
            if ( newcon && oldcon && ((int)(newcon-newpath) == (int)(oldcon-oldpath)) )
            {
               newname.name = newpath;
               newname.len = (int)(newcon-newpath);
               newname.hash = 0;

               oldname.name = oldpath;
               oldname.len = (int)(oldcon-oldpath);
               oldname.hash = 0;
               if ( !Novfs_d_strcmp(&newname, &oldname))
               {

                  if ( od->d_inode && od->d_inode->u.generic_ip )
                  {

                     if (nd->d_inode && nd->d_inode->u.generic_ip)
                     {
                        session = Scope_Get_SessionId(((PINODE_DATA)ndir->u.generic_ip)->Scope);
                        if (0 == session)
                        {
                           ((PINODE_DATA)ndir->u.generic_ip)->Scope = Scope_Get_ScopefromPath( nd );
                           session = Scope_Get_SessionId(((PINODE_DATA)ndir->u.generic_ip)->Scope);
                        }

                        retCode = Novfs_Delete(newpath, S_ISDIR(nd->d_inode->i_mode), session);
                     }


                     session = Scope_Get_SessionId(((PINODE_DATA)ndir->u.generic_ip)->Scope);
                     if (0 == session)
                     {
                        ((PINODE_DATA)ndir->u.generic_ip)->Scope = Scope_Get_ScopefromPath( nd );
                        session = Scope_Get_SessionId(((PINODE_DATA)ndir->u.generic_ip)->Scope);
                     }
                     retCode = Novfs_Rename_File(
                                 S_ISDIR(od->d_inode->i_mode),
                                 oldpath,
                                 oldlen-1,
                                 newpath,
                                 newlen-1, 
                                 session);

                     if ( !retCode )
                     {
                        info = (PENTRY_INFO)oldbuf;
                        od->d_time = 0;
                        Novfs_remove_inode_entry(odir, &od->d_name, 0);
                        Novfs_Get_File_Info(newpath, info, session);
                        nd->d_time = jiffies+(File_update_timeout*HZ);
                        if (Novfs_update_entry(ndir, &nd->d_name, 0, info))
                        {
                           if (nd->d_inode && nd->d_inode->i_ino)
                           {
                              ino = nd->d_inode->i_ino;
                           }
                           else
                           {
                              ino = (ino_t)InterlockedIncrement(&Novfs_Inode_Number);
                           }
                           Novfs_add_inode_entry(ndir, &nd->d_name, ino, info);
                        }
                     }
                  }
               }
            }
         }
      }
   }

   if (oldbuf) Novfs_Free(oldbuf);

   return( retCode );
}

/*++======================================================================*/
int Novfs_i_permission(struct inode *inode, int mask)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode=0;

   return(retCode);
}

/*++======================================================================*/
int Novfs_i_setattr(struct dentry *dentry, struct iattr *attr)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   char *path, *buf;
   struct inode *inode=dentry->d_inode;
   char atime_buf[32];
   char mtime_buf[32];
   char ctime_buf[32];
   unsigned int ia_valid = attr->ia_valid;
   session_t session;
   int retVal = 0;

   if ( IS_ROOT(dentry) || /* Root */
        IS_ROOT(dentry->d_parent) || /* User */
        IS_ROOT(dentry->d_parent->d_parent) || /* Server */
        IS_ROOT(dentry->d_parent->d_parent->d_parent) ) /* Volume */
   {
      return(-EACCES);
   }

   if (inode && inode->u.generic_ip)
   {
      session = Scope_Get_SessionId( ((PINODE_DATA)inode->u.generic_ip)->Scope);
      if (0 == session)
      {
         ((PINODE_DATA)inode->u.generic_ip)->Scope = Scope_Get_ScopefromPath( dentry );
         session = Scope_Get_SessionId(((PINODE_DATA)inode->u.generic_ip)->Scope);
      }

      buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
      if (buf)
      {
         path = Novfs_dget_path(dentry, buf, PATH_LENGTH_BUFFER);
         if (path)
         {
            strcpy(atime_buf, "Unspecified");
            strcpy(mtime_buf, "Unspecified");
            strcpy(ctime_buf, "Unspecified");
            if (attr->ia_valid & ATTR_ATIME)
            {
               ctime_r(&attr->ia_atime.tv_sec, atime_buf);
            }
            if (attr->ia_valid & ATTR_MTIME)
            {
               ctime_r(&attr->ia_mtime.tv_sec, mtime_buf);
            }
            if (attr->ia_valid & ATTR_CTIME)
            {
               ctime_r(&attr->ia_ctime.tv_sec, ctime_buf);
            }
            /* Removed for Bug 132374. jlt
            DbgPrint("Novfs_i_setattr: %s\n" \
                     "   ia_valid:      0x%x\n" \
                     "   ia_mode:       0%o\n" \
                     "   ia_uid:        %d\n" \
                     "   ia_gid:        %d\n" \
                     "   ia_size:       %lld\n" \
                     "   ia_atime:      %s\n" \
                     "   ia_mtime:      %s\n" \
                     "   ia_ctime:      %s\n" \
                     "   ia_attr_flags: 0x%x\n",
               path,
               attr->ia_valid,
               attr->ia_mode,
               attr->ia_uid,
               attr->ia_gid,
               attr->ia_size,
               atime_buf,
               mtime_buf,
               ctime_buf,
               attr->ia_attr_flags);
            */
            
            if ( !(retVal = Novfs_Set_Attr(path, attr, session) ) )
            {
               ((PINODE_DATA)inode->u.generic_ip)->Flags |= UPDATE_INODE;

               if (ia_valid & ATTR_ATIME)
                  inode->i_atime = attr->ia_atime;
               if (ia_valid & ATTR_MTIME)
                  inode->i_mtime = attr->ia_mtime;
               if (ia_valid & ATTR_CTIME)
                  inode->i_ctime = attr->ia_ctime;
               if (ia_valid & ATTR_MODE) {
                  inode->i_mode = attr->ia_mode & (S_IFMT | S_IRWXU);
               }
            }
         }
      }
      Novfs_Free(buf);
   }

   return(retVal);
}

/*++======================================================================*/
int Novfs_i_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *kstat)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode = 0;
   char atime_buf[32];
   char mtime_buf[32];
   char ctime_buf[32];
   struct inode *inode = dentry->d_inode;

   ENTRY_INFO info;
   char *path, *buf;
   session_t session;
   PINODE_DATA id;

   if ( !IS_ROOT(dentry) &&
        !IS_ROOT(dentry->d_parent) )
   {
      session = 0;
      id = dentry->d_inode->u.generic_ip;

      if (id && (id->Flags & UPDATE_INODE) )
      {
         session = Scope_Get_SessionId( id->Scope );

         if (0 == session)
         {
            id->Scope = Scope_Get_ScopefromPath( dentry );
            session = Scope_Get_SessionId(id->Scope);
         }

         buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
         if (buf)
         {
            path = Novfs_dget_path(dentry, buf, PATH_LENGTH_BUFFER);
            if (path)
            {
               retCode = Novfs_Get_File_Info(path, &info, session );
               if ( !retCode )
               {
                  update_inode(inode, &info);
                  id->Flags &= ~UPDATE_INODE;
               }
            }
            Novfs_Free(buf);
         }
      }
   }

   kstat->ino     = inode->i_ino;
   kstat->dev     = inode->i_sb->s_dev;
   kstat->mode    = inode->i_mode;
   kstat->nlink   = inode->i_nlink;
   kstat->uid     = inode->i_uid;
   kstat->gid     = inode->i_gid;
   kstat->rdev    = inode->i_rdev;
   kstat->size    = i_size_read(inode);
   kstat->atime   = inode->i_atime;
   kstat->mtime   = inode->i_mtime;
   kstat->ctime   = inode->i_ctime;
   kstat->blksize = inode->i_blksize;
   kstat->blocks  = inode->i_blocks;
   if (inode->i_bytes)
   {
   	kstat->blocks++;
   }
   ctime_r(&kstat->atime.tv_sec, atime_buf);
   ctime_r(&kstat->mtime.tv_sec, mtime_buf);
   ctime_r(&kstat->ctime.tv_sec, ctime_buf);


   DbgPrint("Novfs_i_getattr: 0x%x <%.*s>\n" \
            "   ino: 0x%x\n" \
            "   dev: 0x%x\n" \
            "   mode: 0%o\n" \
            "   nlink: 0x%x\n" \
            "   uid: 0x%x\n" \
            "   gid: 0x%x\n" \
            "   rdev: 0x%x\n" \
            "   size: 0x%llx\n" \
            "   atime: %s\n" \
            "   mtime: %s\n" \
            "   ctime: %s\n" \
            "   blksize: 0x%x\n" \
            "   blocks: 0x%x\n",
      retCode, dentry->d_name.len, dentry->d_name.name,
      kstat->ino,
      kstat->dev,
      kstat->mode,
      kstat->nlink,
      kstat->uid,
      kstat->gid,
      kstat->rdev,
      kstat->size,
      atime_buf,
      mtime_buf,
      ctime_buf,
      kstat->blksize,
      kstat->blocks);
   return( retCode );
}


/*++======================================================================*/
int Novfs_i_revalidate(struct dentry *dentry)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{

   DbgPrint("Novfs_i_revalidate: name %.*s\n", dentry->d_name.len, dentry->d_name.name);

   return(0);
}

/*++======================================================================*/
void Novfs_read_inode(struct inode *inode)
/*
 *  Arguments:
 *
 *  Returns:      nothing
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:  Superblock operation
 *
 *=======================================================================-*/
{
   DbgPrint( "Novfs_read_inode: 0x%x %d\n", inode, inode->i_ino);
}

/*++======================================================================*/
void Novfs_write_inode(struct inode *inode)
/*
 *  Arguments:
 *
 *  Returns:      nothing
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:  Superblock operation
 *
 *=======================================================================-*/
{
   DbgPrint( "Novfs_write_inode: Inode=0x%x Ino=%d\n", inode, inode->i_ino);
}

/*++======================================================================*/
int Novfs_notify_change(struct dentry *dentry, struct iattr *attr)
/*
 *  Arguments:
 *
 *  Returns:      error code
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:  Superblock operation
 *
 *=======================================================================-*/
{
   struct inode *inode = dentry->d_inode;

   DbgPrint( "Novfs_notify_change: Dentry=0x%x Name=%.*s Inode=0x%x Ino=%d ia_valid=0x%x\n", 
      dentry, dentry->d_name.len, dentry->d_name.name,inode, inode->i_ino, attr->ia_valid);
   return(0);
}

/*++======================================================================*/
void Novfs_clear_inode(struct inode *inode)
/*
 * Arguments:    sb - pointer to the super_block
 *               buf - pointer to the statfs buffer
 *
 *  Returns:      0
 * 
 *  Abstract:     Called when statfs(2) system called.
 * 
 *  Notes:
 * 
 *  Environment:  Superblock operation
 *
 *========================================================================*/
{
   InodeCount--;
   
   if ( inode->u.generic_ip )
   {
      PINODE_DATA id=inode->u.generic_ip;

      DbgPrint("Novfs_clear_inode: inode=0x%x ino=%d Scope=0x%p Name=%s\n", inode, inode->i_ino, id->Scope, id->Name);

      Novfs_free_inode_cache(inode);

      down( &InodeList_lock );
      list_del( &id->IList );
      up( &InodeList_lock );


      Novfs_Free(inode->u.generic_ip);
      inode->u.generic_ip = NULL;

      remove_inode_hash( inode );

   }
   else
   {
      DbgPrint("Novfs_clear_inode: inode=0x%x ino=%d\n", inode, inode->i_ino);
   }
}

/*++======================================================================*/
int 
NO_TRACE
Novfs_show_options( struct seq_file *s, struct vfsmount *m )
/*
 * Arguments:    
 *
 *  Returns:      0
 * 
 *  Abstract:     Called when /proc/mounts is read
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   char *buf, *path, *tmp;

   buf = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER, GFP_KERNEL);
   if (buf)
   {
      path = d_path(m->mnt_root, m, buf, PATH_LENGTH_BUFFER);
      if (path)
      {
         if ( !Novfs_CurrentMount || (Novfs_CurrentMount && strcmp(Novfs_CurrentMount, path)))
         {
            DbgPrint("Novfs_show_options: %.*s %.*s %s\n", m->mnt_root->d_name.len, m->mnt_root->d_name.name, 
               m->mnt_mountpoint->d_name.len, m->mnt_mountpoint->d_name.name, path);
            tmp = (char *)Novfs_Malloc(PATH_LENGTH_BUFFER - (int)(path-buf), GFP_KERNEL);
            if (tmp)
            {
               strcpy(tmp, path);
               path = Novfs_CurrentMount;
               Novfs_CurrentMount = tmp;
               Daemon_SetMountPoint( Novfs_CurrentMount );

               if (path)
               {
                  Novfs_Free(path);
               }
            }
         }
      }
      Novfs_Free( buf );
   }
   return(0);
}

/*++======================================================================*/
int Novfs_statfs(struct super_block *sb, struct kstatfs *buf)
/*
 * Arguments:    sb - pointer to the super_block
 *               buf - pointer to the statfs buffer
 *
 *  Returns:      0
 * 
 *  Abstract:     Called when statfs(2) system called.
 * 
 *  Notes:
 * 
 *  Environment:  Superblock operation
 *
 *========================================================================*/
{
   uint64_t td, fd, te, fe;
 
   UNUSED_VARIABLE(sb);

   DbgPrint("Novfs_statfs:\n");

   td = fd = te = fe = 0;

   Scope_Get_UserSpace( &td, &fd, &te, &fe );

   DbgPrint("td=%llu\n", td);
   DbgPrint("fd=%llu\n", fd);
   DbgPrint("te=%llu\n", te);
   DbgPrint("fe=%llu\n", fd);

   buf->f_type = sb->s_magic;
   buf->f_bsize = sb->s_blocksize;
   buf->f_namelen = NW_MAX_PATH_LENGTH;
   buf->f_blocks = (sector_t)(td + (uint64_t)(sb->s_blocksize-1)) >> (uint64_t)sb->s_blocksize_bits;
   buf->f_bfree = (sector_t)fd >> (uint64_t)sb->s_blocksize_bits;
   buf->f_bavail = (sector_t)buf->f_bfree;
   buf->f_files = (sector_t)te;
   buf->f_ffree = (sector_t)fe;
   buf->f_frsize = sb->s_blocksize;
   if (te > 0xffffffff)
      buf->f_files = 0xffffffff;

   if (fe > 0xffffffff)
      buf->f_ffree = 0xffffffff;


   DbgPrint("f_type:    0x%x\n", buf->f_type);
   DbgPrint("f_bsize:   %u\n", buf->f_bsize);
   DbgPrint("f_namelen: %d\n", buf->f_namelen);
   DbgPrint("f_blocks:  %llu\n", buf->f_blocks);
   DbgPrint("f_bfree:   %llu\n", buf->f_bfree);
   DbgPrint("f_bavail:  %llu\n", buf->f_bavail);
   DbgPrint("f_files:   %llu\n", buf->f_files);
   DbgPrint("f_ffree:   %llu\n", buf->f_ffree);
   DbgPrint("f_frsize:  %u\n", buf->f_frsize);

   return 0;
}

/*++======================================================================*/
struct inode *Novfs_get_inode(struct super_block *sb, int mode, int dev, uid_t Uid, ino_t ino, struct qstr *name)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *=======================================================================-*/
{
   struct inode * inode = new_inode(sb);

   if (inode) 
   {
      InodeCount++;
      inode->i_mode = mode;
      inode->i_uid = Uid;
      inode->i_gid = 0;
      inode->i_blksize = sb->s_blocksize;
      inode->i_blkbits = sb->s_blocksize_bits;
      inode->i_blocks = 0;
      inode->i_rdev = 0;
      inode->i_ino = (ino) ? ino: (ino_t)InterlockedIncrement(&Novfs_Inode_Number);
      inode->i_mapping->a_ops = &Novfs_aops;
      inode->i_atime.tv_sec = 0;
      inode->i_atime.tv_nsec = 0;
      inode->i_mtime = inode->i_ctime = inode->i_atime;

      DbgPrint("Novfs_get_inode: Inode=0x%p I_ino=%d len=%d\n", inode, inode->i_ino, name->len);

      if (NULL != (inode->u.generic_ip = Novfs_Malloc(sizeof(INODE_DATA)+name->len, GFP_KERNEL)))
      {
         PINODE_DATA id;
         id = inode->u.generic_ip;

         DbgPrint("Novfs_get_inode: u.generic_ip 0x%p\n", id);

         id->Scope = NULL;
         id->Flags = 0;
         id->Inode = inode;
         INIT_LIST_HEAD( &id->DirCache );
         init_MUTEX( &id->DirCacheLock );
         down( &InodeList_lock );

         list_add_tail(&id->IList, &InodeList);
         up( &InodeList_lock );

         id->Name[0] = '\0';

         memcpy(id->Name, name->name, name->len);
         id->Name[name->len] = '\0';

         DbgPrint("Novfs_get_inode: name %s\n", id->Name);
      }

		insert_inode_hash(inode);

      switch (mode & S_IFMT) {

      case S_IFREG:
         inode->i_op = &Novfs_inode_operations;
         inode->i_fop = &Novfs_file_operations;
         break;

      case S_IFDIR:
         inode->i_op = &Novfs_inode_operations;
         inode->i_fop = &Novfs_dir_operations;
         inode->i_blksize = 0;
         inode->i_blkbits = 0;
         break;

      default:
         init_special_inode(inode, mode, dev);
         break;
      }

      DbgPrint("Novfs_get_inode: size=%lld\n", inode->i_size);
      DbgPrint("Novfs_get_inode: mode=0%o\n", inode->i_mode);
      DbgPrint("Novfs_get_inode: i_blksize=%d\n", inode->i_blksize);
      DbgPrint("Novfs_get_inode: i_blkbits=%d\n", inode->i_blkbits);
      DbgPrint("Novfs_get_inode: i_blocks=%d\n", inode->i_blocks);
      DbgPrint("Novfs_get_inode: i_bytes=%d\n", inode->i_bytes);
   }

   DbgPrint("Novfs_get_inode: 0x%p %d\n", inode, inode->i_ino);
   return (inode);
}

/*++======================================================================*/
int Novfs_fill_super (struct super_block *SB, void *Data, int Silent)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct inode * inode;
   struct dentry *server, *tree;
   struct qstr name;
/*   PINODE_DATA id;
*/
   ENTRY_INFO info;

   UNUSED_VARIABLE(Data);
   UNUSED_VARIABLE(Silent);

   SB->s_blocksize = PAGE_CACHE_SIZE;
   SB->s_blocksize_bits = PAGE_CACHE_SHIFT;
   SB->s_maxbytes = 0xFFFFFFFFFFFFFFFFULL;	/* Max file size */
   SB->s_op = &Novfs_ops;
   SB->s_flags |= (MS_NODIRATIME | MS_NODEV | MS_POSIXACL);
   SB->s_magic = NOVFS_MAGIC;

   
   name.len = 1;
   name.name = "/";

   inode = Novfs_get_inode(SB, S_IFDIR | 0777, 0, 0, 0, &name);
   if (!inode)
   {
      return( -ENOMEM );
   }
   
   Novfs_root = d_alloc_root(inode);

   if (!Novfs_root)
   {
      iput(inode);
      return( -ENOMEM );
   }
   Novfs_root->d_time = jiffies+(File_update_timeout*HZ);

   inode->i_atime = 
   inode->i_ctime = 
   inode->i_mtime = CURRENT_TIME;


   SB->s_root = Novfs_root;

   DbgPrint( "Novfs_fill_super: root 0x%x\n", Novfs_root);

   if (Novfs_root)
   {
      Novfs_root->d_op = &Novfs_dentry_operations;

      name.name = SERVER_DIRECTORY_NAME;
      name.len  = strlen(SERVER_DIRECTORY_NAME);
      name.hash = Novfs_internal_hash( &name );

      inode = Novfs_get_inode(SB, S_IFDIR | 0777, 0, 0, 0, &name);
      if (inode)
      {
         info.mode = inode->i_mode;
         info.namelength = 0;
         inode->i_size  = info.size = 0;
         inode->i_uid   = info.uid  = 0;
         inode->i_gid   = info.gid  = 0;
         inode->i_atime = info.atime =
         inode->i_ctime = info.ctime =
         inode->i_mtime = info.mtime = CURRENT_TIME;

         server = d_alloc(Novfs_root, &name);
         if (server)
         {
            server->d_op = &Novfs_dentry_operations;
            server->d_time = 0xffffffff;
            d_add(server, inode);
            DbgPrint( "Novfs_fill_super: d_add %s 0x%x\n", SERVER_DIRECTORY_NAME, server);
            Novfs_add_inode_entry(Novfs_root->d_inode, &name, inode->i_ino, &info);
         }
      }

      name.name = TREE_DIRECTORY_NAME;
      name.len  = strlen(TREE_DIRECTORY_NAME);
      name.hash =  Novfs_internal_hash( &name );

      inode = Novfs_get_inode(SB, S_IFDIR | 0777, 0, 0, 0, &name);
      if (inode)
      {
         info.mode = inode->i_mode;
         info.namelength = 0;
         inode->i_size  = info.size = 0;
         inode->i_uid   = info.uid  = 0;
         inode->i_gid   = info.gid  = 0;
         inode->i_atime = info.atime =
         inode->i_ctime = info.ctime =
         inode->i_mtime = info.mtime = CURRENT_TIME;
         tree = d_alloc(Novfs_root, &name);
         if (tree)
         {
            tree->d_op = &Novfs_dentry_operations;
            tree->d_time = 0xffffffff;

            d_add(tree, inode);
            DbgPrint( "Novfs_fill_super: d_add %s 0x%x\n", TREE_DIRECTORY_NAME, tree);
            Novfs_add_inode_entry(Novfs_root->d_inode, &name, inode->i_ino, &info);
         }
      }
   }

   return( 0 );
}

/*++======================================================================*/
struct super_block *Novfs_get_sb(struct file_system_type *Fstype, int Flags, const char *Dev_name, void *Data)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   struct super_block *sb;
   UNUSED_VARIABLE(Dev_name);

   sb = get_sb_nodev(Fstype, Flags, Data, Novfs_fill_super);

   DbgPrint( "Novfs_get_sb: sb=0x%x Fstype=0x%x Dev_name=%s\n", sb, Fstype, Dev_name);

   return (sb );
}

/*++======================================================================*/
void Novfs_kill_sb(struct super_block *SB)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   kill_litter_super(SB);
}

/*++======================================================================*/
ssize_t Novfs_Control_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   ssize_t   retval=0;

   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(buf);
   UNUSED_VARIABLE(nbytes);
   UNUSED_VARIABLE(ppos);

   DbgPrint( "Novfs_Control_read: kernel_locked 0x%x\n", kernel_locked());

   return retval;
}

/*++======================================================================*/
ssize_t Novfs_Control_write(struct file * file, const char * buf, size_t nbytes, loff_t *ppos)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   ssize_t   retval=0;

   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(buf);
   UNUSED_VARIABLE(nbytes);
   UNUSED_VARIABLE(ppos);

   DbgPrint( "Novfs_Control_write: kernel_locked 0x%x\n", kernel_locked());
   if (buf && nbytes)
   {
   }
   
   return(retval);
}

/*++======================================================================*/
int Novfs_Control_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retval=0;

   UNUSED_VARIABLE(inode);
   UNUSED_VARIABLE(file);
   UNUSED_VARIABLE(cmd);
   UNUSED_VARIABLE(arg);

   DbgPrint( "Novfs_Control_ioctl: kernel_locked 0x%x\n", kernel_locked());

   return(retval);
}

/*++======================================================================*/
int __init init_novfs (void)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   int retCode;
   

   retCode = Init_Procfs_Interface();

   init_profile();

   if ( !retCode )
   {
      DbgPrint("init_novfs: %s %s %s\n", __DATE__, __TIME__, NOVFS_VERSION_STRING);
      Init_Daemon_Queue();
      Scope_Init();
      retCode = register_filesystem(&Novfs_fs_type);
      if ( retCode )
      {
         Uninit_Procfs_Interface();
         Uninit_Daemon_Queue();
         Scope_Uninit();
      }
   }
   return(retCode);
}

/*++======================================================================*/
void __exit exit_novfs(void)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:
 * 
 *  Environment:
 *
 *========================================================================*/
{
   printk( KERN_INFO "exit_novfs\n");
   Scope_Uninit();
   printk( KERN_INFO "exit_novfs after Scope_Uninit\n");
   Uninit_Daemon_Queue();
   printk( KERN_INFO "exit_novfs after Uninit_Daemon_Queue\n");
   Uninit_Procfs_Interface();
   printk( KERN_INFO "exit_novfs Uninit_Procfs_Interface\n");
   unregister_filesystem(&Novfs_fs_type);
   printk( KERN_INFO "exit_novfs: Exit\n");
   if (Novfs_CurrentMount)
   {
      Novfs_Free(Novfs_CurrentMount);
      Novfs_CurrentMount = NULL;
   }
}

/*++======================================================================*/
int Novfs_lock_inode_cache( struct inode *i )
/*
 *
 *  Arguments:   struct inode *i - pointer to directory inode
 *
 *  Returns:     0 - locked
 *              -1 - not locked
 *
 *  Abstract:    Locks the inode cache.
 *
 *  Notes:       
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   int retVal = 0;

   DbgPrint("Novfs_lock_inode_cache: 0x%p\n", i);
   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      down( &id->DirCacheLock );
      retVal = 1;
   }
   DbgPrint("Novfs_lock_inode_cache: return %d\n", retVal);
   return( retVal );
}

/*++======================================================================*/
void Novfs_unlock_inode_cache( struct inode *i )
/*
 *  Arguments:   struct inode *i - pointer to directory inode
 *
 *  Returns:     nothing
 * 
 *  Abstract:    Unlocks inode cache.
 * 
 *  Notes:       
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      up( &id->DirCacheLock );
   }
}

/*++======================================================================*/
int Novfs_enumerate_inode_cache( struct inode *i, struct list_head **iteration, ino_t *ino, PENTRY_INFO info)
/*
 *  Arguments:   struct inode *i - pointer to directory inode
 *
 *  Returns:     0 - item found
 *              -1 - done
 *
 *  Abstract:    Unlocks inode cache.
 *
 *  Notes:       DirCacheLock should be held before calling this routine.
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *l=NULL;
   int retVal = -1;


   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      if ( (NULL == iteration) || (NULL == *iteration) )
      {
         l = id->DirCache.next;
      }
      else
      {
         l = *iteration;
      }

      if (l == &id->DirCache)
      {
         l = NULL;
      }
      else
      {
         dc = list_entry(l, DIR_CACHE, list);

         *ino = dc->ino;
         info->type = 0;
         info->mode = dc->mode;
         info->size = dc->size;
         info->atime = dc->atime;
         info->mtime = dc->mtime;
         info->ctime = dc->ctime;
         info->namelength = dc->nameLen;
         memcpy(info->name, dc->name, dc->nameLen);
         info->name[dc->nameLen] = '\0';
         retVal = 0;

         l = l->next;
      }
   }
   *iteration = l;
   return( retVal );
}

/*++======================================================================*/
int Novfs_get_entry( struct inode *i, struct qstr *name, ino_t *ino, PENTRY_INFO info)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:       DirCacheLock should be held before calling this routine.
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   int retVal = -1;
   char *n="<NULL>";
   int nl=6;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      if (name && name->len)
      {
         n = (char *)name->name;
         nl = name->len;
      }
      DbgPrint("Novfs_get_entry:\n" \
               "   inode: 0x%p\n" \
               "   name:  %.*s\n" \
               "   ino:   %d\n",
         i, nl, n, *ino);

      dc = Novfs_lookup_inode_cache(i, name, *ino);
      if (dc)
      {
         dc->flags |= ENTRY_VALID;
         retVal = 0;
         *ino = dc->ino;
         info->type = 0;
         info->mode = dc->mode;
         info->size = dc->size;
         info->atime = dc->atime;
         info->mtime = dc->mtime;
         info->ctime = dc->ctime;
         info->namelength = dc->nameLen;
         memcpy(info->name, dc->name, dc->nameLen);
         info->name[dc->nameLen] = '\0';
         retVal = 0;
      }
   }
   DbgPrint("Novfs_get_entry: return %d\n", retVal);
   return( retVal );
}

/*++======================================================================*/
int Novfs_get_entry_time( struct inode *i, struct qstr *name, ino_t *ino, PENTRY_INFO info, u64 *EntryTime)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:       DirCacheLock should be held before calling this routine.
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   int retVal = -1;
   char *n="<NULL>";
   int nl=6;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      if (name && name->len)
      {
         n = (char *)name->name;
         nl = name->len;
      }
      DbgPrint("Novfs_get_entry:\n" \
               "   inode: 0x%p\n" \
               "   name:  %.*s\n" \
               "   ino:   %d\n",
         i, nl, n, *ino);

      dc = Novfs_lookup_inode_cache(i, name, *ino);
      if (dc)
      {
         retVal = 0;
         *ino = dc->ino;
         info->type = 0;
         info->mode = dc->mode;
         info->size = dc->size;
         info->atime = dc->atime;
         info->mtime = dc->mtime;
         info->ctime = dc->ctime;
         info->namelength = dc->nameLen;
         memcpy(info->name, dc->name, dc->nameLen);
         info->name[dc->nameLen] = '\0';
         if (EntryTime)
         {
            *EntryTime = dc->jiffies;
         }
         retVal = 0;
      }
   }
   DbgPrint("Novfs_get_entry: return %d\n", retVal);
   return( retVal );
}

/*++======================================================================*/
int Novfs_get_remove_entry( struct inode *i, ino_t *ino, PENTRY_INFO info)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:    This routine will return the first entry on the list 
 *               and then remove it.
 *
 *  Notes:       DirCacheLock should be held before calling this routine.
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *l=NULL;
   int retVal = -1;


   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      l = id->DirCache.next;

      if (l != &id->DirCache)
      {
         dc = list_entry(l, DIR_CACHE, list);

         *ino = dc->ino;
         info->type = 0;
         info->mode = dc->mode;
         info->size = dc->size;
         info->atime = dc->atime;
         info->mtime = dc->mtime;
         info->ctime = dc->ctime;
         info->namelength = dc->nameLen;
         memcpy(info->name, dc->name, dc->nameLen);
         info->name[dc->nameLen] = '\0';
         retVal = 0;

         list_del( &dc->list );
         Novfs_Free( dc );
         DCCount--;

      }
   }
   return( retVal );
}

/*++======================================================================*/
void Novfs_invalidate_inode_cache( struct inode *i )
/*
 *  Arguments:   struct inode *i - pointer to directory inode
 *
 *  Returns:     nothing
 * 
 *  Abstract:    Marks all entries in the directory cache as invalid.
 * 
 *  Notes:       DirCacheLock should be held before calling this routine.
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *l;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      list_for_each(l, &id->DirCache)
      {
         dc = list_entry(l, DIR_CACHE, list);
         dc->flags &= ~ENTRY_VALID;
      }
   }
}

/*++======================================================================*/
PDIR_CACHE Novfs_lookup_inode_cache( struct inode *i, struct qstr *name, ino_t ino )
/*
 *  Arguments:   struct inode *i - pointer to directory inode
 *               struct qstr *name - pointer to name
 *               ino_t - inode number
 *
 *  Returns:     DIR_CACHE entry if match
 *               NULL - if there is no match.
 *
 *  Abstract:    Checks a inode directory to see if there are any enties
 *               matching name or ino.  If name is specified then ino is 
 *               not used.  ino is use if name is not specified.
 *
 *  Notes:       DirCacheLock should be held before calling this routine.
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc, retVal=NULL;
   struct list_head *l;
   char *n="<NULL>";
   int nl=6;
   int hash=0;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      if (name && name->name)
      {
         nl = name->len;
         n = (char *)name->name;
         hash = name->hash;
      }
      DbgPrint("Novfs_lookup_inode_cache:\n" \
               "   inode: 0x%p\n" \
               "   name:  %.*s\n" \
               "   hash:  0x%x\n" \
               "   len:   %d\n" \
               "   ino:   %d\n",
         i, nl, n, hash, nl, ino);

      list_for_each(l, &id->DirCache)
      {
         dc = list_entry(l, DIR_CACHE, list);
         if (name)
         {
            
/*         DbgPrint("Novfs_lookup_inode_cache: 0x%p\n" \
                  "   ino:   %d\n" \
                  "   hash:  0x%x\n" \
                  "   len:   %d\n" \
                  "   name:  %.*s\n",
            dc, dc->ino, dc->hash, dc->nameLen, dc->nameLen, dc->name);
*/
            if ( (name->hash == dc->hash) &&
                 (name->len == dc->nameLen) &&
                 (0 == memcmp(name->name, dc->name, name->len)) )
            {
               retVal = dc;
               break;
            }
         }
         else
         {
            if (ino == dc->ino)
            {
               retVal = dc;
               break;
            }
         }
      }
   }

   DbgPrint("Novfs_lookup_inode_cache: return 0x%p\n", retVal);
   return( retVal );
}

/*++======================================================================*/
int Novfs_lookup_validate( struct inode *i, struct qstr *name, ino_t ino )
/*
 *  Arguments:   struct inode *i - pointer to directory inode
 *               struct qstr *name - pointer to name
 *               ino_t - inode number
 *
 *  Returns:     0 if found
 *               !0 if not found
 *
 *  Abstract:    Checks a inode directory to see if there are any enties
 *               matching name or ino.  If entry is found the valid bit
 *               is set.
 *
 *  Notes:       DirCacheLock should be held before calling this routine.
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   int retVal = -1;
   char *n="<NULL>";
   int nl=6;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      if (name && name->len)
      {
         n = (char *)name->name;
         nl = name->len;
      }
      DbgPrint("Novfs_update_entry:\n" \
               "   inode: 0x%p\n" \
               "   name:  %.*s\n" \
               "   ino:   %d\n",
         i, nl, n, ino);

      dc = Novfs_lookup_inode_cache( i, name, ino );
      if (dc)
      {
         dc->flags |= ENTRY_VALID;
         retVal = 0;
      }
   }
   return( retVal );
}

/*++======================================================================*/
int Novfs_add_inode_entry( 
   struct inode    *i, 
   struct qstr     *name, 
   ino_t            ino, 
   PENTRY_INFO      info)
/*
 *  Arguments:
 *
 *  Returns:     -ENOMEM - alloc error.
 *                0 - success.
 *
 *  Abstract:    Added entry to directory cache.
 *
 *  Notes:       DirCacheLock should be held before calling this routine.
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE new;
   int retVal = -ENOMEM;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      new = Novfs_Malloc(sizeof(DIR_CACHE)+name->len, GFP_KERNEL);
      if (new)
      {
         DCCount++;
         DbgPrint("Novfs_add_inode_entry:\n" \
                  "   inode: 0x%p\n" \
                  "   id:    0x%p\n" \
                  "   DC:    0x%p\n" \
                  "   new:   0x%p\n" \
                  "   name:  %.*s\n" \
                  "   ino:   %d\n"   \
                  "   size:  %lld\n" \
                  "   mode:  0x%x\n",
            i, id, &id->DirCache, new, name->len, name->name, ino, info->size, info->mode);

         retVal = 0;
         new->flags   = ENTRY_VALID;
         new->jiffies = get_jiffies_64();
         new->size    = info->size;
         new->mode    = info->mode;
         new->atime   = info->atime;
         new->mtime   = info->mtime;
         new->ctime   = info->ctime;
         new->ino     = ino;
         new->hash    = name->hash;
         new->nameLen = name->len;
         memcpy(new->name, name->name, name->len);
         new->name[new->nameLen] = '\0';
         list_add(&new->list, &id->DirCache);

/*         list_for_each(l, &id->DirCache)
         {
            dc = list_entry(l, DIR_CACHE, list);
            if ( dc->hash > new->hash )
            {
               break;
            }
         }

         DbgPrint("Novfs_add_inode_entry: adding 0x%p to 0x%p\n", new, l);
         list_add(&new->list, l);
*/
      }
   }
   return( retVal );
}

/*++======================================================================*/
int Novfs_update_entry( struct inode *i, struct qstr *name, ino_t ino, PENTRY_INFO info)
/*
 *  Arguments:
 *
 *  Returns:
 * 
 *  Abstract:
 * 
 *  Notes:       DirCacheLock should be held before calling this routine.
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   int retVal = -1;
   char *n="<NULL>";
   int nl=6;
   char atime_buf[32];
   char mtime_buf[32];
   char ctime_buf[32];

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {

      if (name && name->len)
      {
         n = (char *)name->name;
         nl = name->len;
      }
      ctime_r(&info->atime.tv_sec, atime_buf);
      ctime_r(&info->mtime.tv_sec, mtime_buf);
      ctime_r(&info->ctime.tv_sec, ctime_buf);
      DbgPrint("Novfs_update_entry:\n" \
               "   inode: 0x%p\n" \
               "   name:  %.*s\n" \
               "   ino:   %d\n" \
               "   size:  %lld\n" \
               "   atime: %s\n" \
               "   mtime: %s\n" \
               "   ctime: %s\n",
         i, nl, n, ino, info->size, atime_buf, mtime_buf, ctime_buf);


      dc = Novfs_lookup_inode_cache(i, name, ino);
      if (dc)
      {
         retVal = 0;
         dc->flags   = ENTRY_VALID;
         dc->jiffies = get_jiffies_64();
         dc->size    = info->size;
         dc->mode    = info->mode;
         dc->atime   = info->atime;
         dc->mtime   = info->mtime;
         dc->ctime   = info->ctime;

         ctime_r(&dc->atime.tv_sec, atime_buf);
         ctime_r(&dc->mtime.tv_sec, mtime_buf);
         ctime_r(&dc->ctime.tv_sec, ctime_buf);
         DbgPrint("Novfs_update_entry entry: 0x%x\n" \
                  "   flags:   0x%x\n" \
                  "   jiffies: %lld\n" \
                  "   ino:     %d\n"   \
                  "   size:    %lld\n" \
                  "   mode:    0%o\n"  \
                  "   atime:   %s\n"   \
                  "   mtime:   %s\n"   \
                  "   ctime:   %s\n"   \
                  "   hash:    0x%x\n" \
                  "   nameLen: %d\n"   \
                  "   name:    %s\n",
            dc, dc->flags, dc->jiffies, dc->ino, dc->size, dc->mode,
            atime_buf, mtime_buf, ctime_buf, dc->hash, dc->nameLen, dc->name);
      }
   }
   DbgPrint("Novfs_update_entry: return %d\n", retVal);
   return( retVal );
}

/*++======================================================================*/
void Novfs_remove_inode_entry( struct inode *i, struct qstr *name, ino_t ino)
/*
 *  Arguments:
 *
 *  Returns:     nothing
 *
 *  Abstract:    Removes entry from directory cache.  You can specify a name
 *               or an inode number.
 *
 *  Notes:       DirCacheLock should be held before calling this routine.
 *
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   char *n="<NULL>";
   int nl=6;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      dc = Novfs_lookup_inode_cache( i, name, ino );
      if (dc)
      {
         if (name && name->name)
         {
            nl = name->len;
            n = (char *)name->name;
         }
         DbgPrint("Novfs_remove_inode_entry:\n" \
                  "   inode: 0x%p\n" \
                  "   id:    0x%p\n" \
                  "   DC:    0x%p\n" \
                  "   name:  %.*s\n" \
                  "   ino:   %d\n" \
                  "   entry: 0x%p\n" \
                  "      name: %.*s\n"\
                  "      ino:  %d\n" \
                  "      next: 0x%p\n" \
                  "      prev: 0x%p\n",
            i, id, &id->DirCache, nl, n, ino, dc, dc->nameLen, dc->name, dc->ino, dc->list.next, dc->list.prev);
         list_del( &dc->list );
         Novfs_Free( dc );
         DCCount--;

      }
   }
}

/*++======================================================================*/
void Novfs_free_invalid_entries( struct inode *i )
/*
 *  Arguments:   struct inode *i - pointer to directory inode.
 *
 *  Returns:     nothing
 * 
 *  Abstract:    Frees all invalid entries in the directory cache.
 * 
 *  Notes:       DirCacheLock should be held before calling this routine.
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *l;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      list_for_each( l, &id->DirCache )
      {
         dc = list_entry( l, DIR_CACHE, list );
         if ( 0 == (dc->flags & ENTRY_VALID) )
         {
            DbgPrint("Novfs_free_invalid_entries:\n" \
                     "   inode: 0x%p\n" \
                     "   id:    0x%p\n" \
                     "   entry:    0x%p\n" \
                     "   name:  %.*s\n" \
                     "   ino:   %d\n",
               i, id, dc, dc->nameLen, dc->name, dc->ino);
            l = l->prev;
            list_del( &dc->list );
            Novfs_Free( dc );
            DCCount--;
         }
      }
   }
}

/*++======================================================================*/
void Novfs_free_inode_cache( struct inode *i )
/*
 *  Arguments:   struct inode *i - pointer to directory inode.
 *
 *  Returns:     nothing
 * 
 *  Abstract:    Frees all entries in the inode cache.
 * 
 *  Notes:       DirCacheLock should be held before calling this routine.
 * 
 *  Environment:
 *
 *========================================================================*/
{
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *l;

   if ( i && (id = i->u.generic_ip) && id->DirCache.next )
   {
      list_for_each( l, &id->DirCache )
      {
         dc = list_entry( l, DIR_CACHE, list );
         l = l->prev;
         list_del( &dc->list );
         Novfs_Free( dc );
         DCCount--;
      }
   }
}

/*++======================================================================*/
int
NO_TRACE
Novfs_dump_inode_cache(int argc, const char **argv, const char **envp, struct pt_regs *regs)
/*
 * Arguments:
 * 
 *  Returns:
 *  
 *  Abstract:
 *  
 *  Notes:
 *  
 *  Environment:
 * 
 *========================================================================*/
{
#ifdef	CONFIG_KDB
   struct inode *inode=NULL;
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *l;
   char atime_buf[32];
   char mtime_buf[32];
   char ctime_buf[32];

   if (Novfs_root)
   {
      inode = Novfs_root->d_inode;
   }

   if (argc > 0)
   {
      inode = (void *)simple_strtoul(argv[1], NULL, 0);
   }

   kdb_printf("Inode: 0x%p\n", inode);
   if (inode)
   {
      id = inode->u.generic_ip;
      kdb_printf("INODE_DATA: 0x%p\n", id);

      if ( id && id->DirCache.next )
      {
         list_for_each(l, &id->DirCache)
         {
            dc = list_entry(l, DIR_CACHE, list);
            ctime_r(&dc->atime.tv_sec, atime_buf);
            ctime_r(&dc->mtime.tv_sec, mtime_buf);
            ctime_r(&dc->ctime.tv_sec, ctime_buf);

            DbgPrint("Cache Entry: 0x%p\n" \
                     "   flags:   0x%x\n" \
                     "   jiffies: %llu\n" \
                     "   ino:     %u\n" \
                     "   size:    %llu\n" \
                     "   mode:    0%o\n" \
                     "   atime:   %s\n" \
                     "   mtime:   %s\n" \
                     "   ctime:   %s\n" \
                     "   hash:    0x%x\n" \
                     "   len:     %d\n" \
                     "   name:    %s\n",
               dc, dc->flags, dc->jiffies,
               dc->ino, dc->size, dc->mode,
               atime_buf, mtime_buf, ctime_buf,
               dc->hash, dc->nameLen, dc->name);
         }
      }
   }
#endif
   return(0);
}

/*++======================================================================*/
void
NO_TRACE
Novfs_dump_inode( void *pf )
/*
 * Arguments:
 * 
 *  Returns:
 *  
 *  Abstract:
 *  
 *  Notes:
 *  
 *  Environment:
 * 
 *========================================================================*/
{
   struct inode *inode;
   void (*pfunc)(char *Fmt, ...) = pf;
   PINODE_DATA id;
   PDIR_CACHE dc;
   struct list_head *il, *l;
   char atime_buf[32];
   char mtime_buf[32];
   char ctime_buf[32];
   unsigned long icnt=0, dccnt=0;

   down( &InodeList_lock );
   list_for_each(il, &InodeList)
   {
      id  = list_entry(il, INODE_DATA, IList);
      inode = id->Inode;
      if (inode)
      {
         icnt++;

         pfunc("Inode=0x%x I_ino=%d\n", inode, inode->i_ino);

         pfunc("   atime=%s\n", ctime_r(&inode->i_atime.tv_sec, atime_buf));
         pfunc("   ctime=%s\n", ctime_r(&inode->i_mtime.tv_sec, atime_buf));
         pfunc("   mtime=%s\n", ctime_r(&inode->i_ctime.tv_sec, atime_buf));
         pfunc("   size=%lld\n", inode->i_size);
         pfunc("   mode=0%o\n", inode->i_mode);
      }

      pfunc("   INODE_DATA: 0x%p Name=%s Scope=0x%p\n", id, id->Name, id->Scope);

      if (id->DirCache.next )
      {
         list_for_each(l, &id->DirCache)
         {
            dccnt++;
            dc = list_entry(l, DIR_CACHE, list);
            ctime_r(&dc->atime.tv_sec, atime_buf);
            ctime_r(&dc->mtime.tv_sec, mtime_buf);
            ctime_r(&dc->ctime.tv_sec, ctime_buf);

            pfunc("   Cache Entry: 0x%p\n" \
                       "      flags:   0x%x\n" \
                       "      jiffies: %llu\n" \
                       "      ino:     %u\n" \
                       "      size:    %llu\n" \
                       "      mode:    0%o\n" \
                       "      atime:   %s\n" \
                       "      mtime:   %s\n" \
                       "      ctime:   %s\n" \
                       "      hash:    0x%x\n" \
                       "      len:     %d\n" \
                       "      name:    %s\n",
               dc, dc->flags, dc->jiffies,
               dc->ino, dc->size, dc->mode,
               atime_buf, mtime_buf, ctime_buf,
               dc->hash, dc->nameLen, dc->name);
         }
      }
   }
   up( &InodeList_lock );

   pfunc("Inodes: %d(%d) DirCache: %d(%d)\n", InodeCount, icnt, DCCount, dccnt );
   
}

module_init(init_novfs)
module_exit(exit_novfs)

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Novell Inc.");
MODULE_DESCRIPTION("Novell NetWare Client for Linux");
MODULE_VERSION( NOVFS_VERSION_STRING );
